# Copyright 2018 FUJITSU LIMITED
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
Test class for IRMC BIOS configuration
"""

from unittest import mock

from ironic.common import exception
from ironic.conductor import task_manager
from ironic.drivers.modules.irmc import bios as irmc_bios
from ironic.drivers.modules.irmc import common as irmc_common
from ironic import objects
from ironic.tests.unit.drivers.modules.irmc import test_common


class IRMCBIOSTestCase(test_common.BaseIRMCTest):

    def setUp(self):
        super(IRMCBIOSTestCase, self).setUp()

    @mock.patch.object(irmc_common, 'parse_driver_info',
                       autospec=True)
    def test_validate(self, parse_driver_info_mock):
        with task_manager.acquire(self.context, self.node.uuid) as task:
            task.driver.bios.validate(task)
            parse_driver_info_mock.assert_called_once_with(task.node)

    def test_apply_configuration(self):
        if not mock._is_instance_mock(irmc_bios.irmc.elcm):
            mock.patch.object(irmc_bios.irmc, 'elcm', autospec=True).start()
        set_bios_configuration_mock = (
            irmc_bios.irmc.elcm.set_bios_configuration)
        get_bios_settings_mock = irmc_bios.irmc.elcm.get_bios_settings
        settings = [{
            "name": "launch_csm_enabled",
            "value": True
        }, {
            "name": "hyper_threading_enabled",
            "value": True
        }, {
            "name": "cpu_vt_enabled",
            "value": True
        }]
        with task_manager.acquire(self.context, self.node.uuid) as task:
            irmc_info = irmc_common.parse_driver_info(task.node)
            task.node.save = mock.Mock()
            get_bios_settings_mock.return_value = settings
            task.driver.bios.apply_configuration(task, settings)
            set_bios_configuration_mock.assert_called_once_with(irmc_info,
                                                                settings)

    def test_apply_configuration_failed(self):
        if not mock._is_instance_mock(irmc_bios.irmc.elcm):
            mock.patch.object(irmc_bios.irmc, 'elcm', autospec=True).start()
        set_bios_configuration_mock = (
            irmc_bios.irmc.elcm.set_bios_configuration)
        settings = [{
            "name": "launch_csm_enabled",
            "value": True
        }, {
            "name": "hyper_threading_enabled",
            "value": True
        }, {
            "name": "setting",
            "value": True
        }]
        irmc_bios.irmc.scci.SCCIError = Exception
        set_bios_configuration_mock.side_effect = Exception
        with task_manager.acquire(self.context, self.node.uuid) as task:
            self.assertRaises(exception.IRMCOperationError,
                              task.driver.bios.apply_configuration,
                              task, settings)

    def test_factory_reset(self):
        with task_manager.acquire(self.context, self.node.uuid) as task:
            self.assertRaises(exception.UnsupportedDriverExtension,
                              task.driver.bios.factory_reset, task)

    @mock.patch.object(objects.BIOSSettingList, 'sync_node_setting',
                       autospec=True)
    @mock.patch.object(objects.BIOSSettingList, 'create',
                       autospec=True)
    @mock.patch.object(objects.BIOSSettingList, 'save',
                       autospec=True)
    @mock.patch.object(objects.BIOSSettingList, 'delete',
                       autospec=True)
    def test_cache_bios_settings(self, delete_mock, save_mock, create_mock,
                                 sync_node_setting_mock):
        if not mock._is_instance_mock(irmc_bios.irmc.elcm):
            mock.patch.object(irmc_bios.irmc, 'elcm', autospec=True).start()
        get_bios_settings_mock = irmc_bios.irmc.elcm.get_bios_settings
        settings = [{
            "name": "launch_csm_enabled",
            "value": True
        }, {
            "name": "hyper_threading_enabled",
            "value": True
        }, {
            "name": "cpu_vt_enabled",
            "value": True
        }]
        with task_manager.acquire(self.context, self.node.uuid) as task:
            irmc_info = irmc_common.parse_driver_info(task.node)
            get_bios_settings_mock.return_value = settings
            sync_node_setting_mock.return_value = \
                (
                    [
                        {
                            "name": "launch_csm_enabled",
                            "value": True
                        }],
                    [
                        {
                            "name": "hyper_threading_enabled",
                            "value": True
                        }],
                    [
                        {
                            "name": "cpu_vt_enabled",
                            "value": True
                        }],
                    []
                )
            task.driver.bios.cache_bios_settings(task)
            get_bios_settings_mock.assert_called_with(irmc_info)
            sync_node_setting_mock.assert_called_once_with(task.context,
                                                           task.node.id,
                                                           settings)
            create_mock.assert_called_once_with(
                task.context, task.node.id,
                sync_node_setting_mock.return_value[0])
            save_mock.assert_called_once_with(
                task.context, task.node.id,
                sync_node_setting_mock.return_value[1])
            delete_names = \
                [setting['name'] for setting in
                 sync_node_setting_mock.return_value[2]]
            delete_mock.assert_called_once_with(task.context, task.node.id,
                                                delete_names)

    def test_cache_bios_settings_failed(self):
        if not mock._is_instance_mock(irmc_bios.irmc.elcm):
            mock.patch.object(irmc_bios.irmc, 'elcm', autospec=True).start()
        get_bios_settings_mock = irmc_bios.irmc.elcm.get_bios_settings
        irmc_bios.irmc.scci.SCCIError = Exception
        get_bios_settings_mock.side_effect = Exception
        with task_manager.acquire(self.context, self.node.uuid) as task:
            self.assertRaises(exception.IRMCOperationError,
                              task.driver.bios.cache_bios_settings,
                              task)



# Copyright 2022 FUJITSU LIMITED
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
Vendor interface of iRMC driver
"""

from ironic.common import exception
from ironic.common.i18n import _
from ironic.drivers import base
from ironic.drivers.modules.irmc import common as irmc_common


class IRMCVendorPassthru(base.VendorInterface):
    def get_properties(self):
        """Return the properties of the interface.

        :returns: Dictionary of <property name>:<property description> entries.
        """
        return irmc_common.COMMON_PROPERTIES

    def validate(self, task, method=None, **kwargs):
        """Validate vendor-specific actions.

        This method validates whether the 'driver_info' property of the
        supplied node contains the required information for this driver.

        :param task: An instance of TaskManager.
        :param method: Name of vendor passthru method
        :raises: InvalidParameterValue if invalid value is contained
            in the 'driver_info' property.
        :raises: MissingParameterValue if some mandatory key is missing
            in the 'driver_info' property.
        """
        irmc_common.parse_driver_info(task.node)

    @base.passthru(['POST'],
                   async_call=True,
                   description='Connect to iRMC and fetch iRMC firmware '
                   'version and, if firmware version has not been cached '
                   'in or actual firmware version is different from one in '
                   'driver_internal_info/irmc_fw_version, store firmware '
                   'version in driver_internal_info/irmc_fw_version.',
                   attach=False,
                   require_exclusive_lock=False)
    def cache_irmc_firmware_version(self, task, **kwargs):
        """Fetch and save iRMC firmware version.

        This method connects to iRMC and fetch iRMC firmware version.
        If fetched firmware version is not cached in or is different from
        one in driver_internal_info/irmc_fw_version, store fetched version
        in driver_internal_info/irmc_fw_version.

        :param task: An instance of TaskManager.
        :raises: IRMCOperationError if some error occurs
        """
        try:
            irmc_common.set_irmc_version(task)
        except (exception.IRMCOperationError,
                exception.InvalidParameterValue,
                exception.MissingParameterValue,
                exception.NodeLocked) as e:
            raise exception.IRMCOperationError(
                operation=_('caching firmware version'), error=e)



# Copyright 2015 FUJITSU LIMITED
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
iRMC Boot Driver
"""

import os
import shutil
import tempfile
from urllib import parse as urlparse

from oslo_log import log as logging
from oslo_utils import importutils
from oslo_utils import netutils

from ironic.common import boot_devices
from ironic.common import exception
from ironic.common.glance_service import service_utils
from ironic.common.i18n import _
from ironic.common import image_service
from ironic.common import images
from ironic.common import metrics_utils
from ironic.common import states
from ironic.common import utils
from ironic.conductor import utils as manager_utils
from ironic.conf import CONF
from ironic.drivers import base
from ironic.drivers.modules import boot_mode_utils
from ironic.drivers.modules import deploy_utils
from ironic.drivers.modules.irmc import common as irmc_common
from ironic.drivers.modules.irmc import management as irmc_management
from ironic.drivers.modules import pxe
from ironic.drivers import utils as driver_utils


scci = importutils.try_import('scciclient.irmc.scci')
viom = importutils.try_import('scciclient.irmc.viom.client')

try:
    if CONF.debug:
        scci.DEBUG = True
except Exception:
    pass

LOG = logging.getLogger(__name__)

METRICS = metrics_utils.get_metrics_logger(__name__)

REQUIRED_PROPERTIES = {
    'deploy_iso': _("Deployment ISO image file name. Required."),
}

RESCUE_PROPERTIES = {
    'rescue_iso': _("UUID (from Glance) of the rescue ISO. Only "
                    "required if rescue mode is being used and ironic "
                    "is managing booting the rescue ramdisk.")
}

OPTIONAL_PROPERTIES = {
    'irmc_pci_physical_ids':
        _("Physical IDs of PCI cards. A dictionary of pairs of resource UUID "
          "and its physical ID like '<UUID>:<Physical ID>,...'. The resources "
          "are Ports and Volume connectors. The Physical ID consists of card "
          "type, slot No, and port No. The format is "
          "{LAN|FC|CNA}<slot-No>-<Port-No>. This parameter is necessary for "
          "booting a node from a remote volume. Optional."),
    'irmc_storage_network_size':
        _("Size of the network for iSCSI storage network. This is the size of "
          "the IPv4 subnet mask that the storage network is configured to "
          "utilize, in a range between 1 and 31 inclusive. This is necessary "
          "for booting a node from a remote iSCSI volume. Optional."),
    'kernel_append_params': driver_utils.KERNEL_APPEND_PARAMS_DESCRIPTION %
    {'option_group': 'irmc'},
}

COMMON_PROPERTIES = REQUIRED_PROPERTIES.copy()
COMMON_PROPERTIES.update(driver_utils.OPTIONAL_PROPERTIES)
COMMON_PROPERTIES.update(OPTIONAL_PROPERTIES)


def _is_image_href_ordinary_file_name(image_href):
    """Check if image_href is an ordinary file name.

    This method judges if image_href is an ordinary file name or not,
    which is a file supposed to be stored in share file system.
    The ordinary file name is neither glance image href
    nor image service href.

    :returns: True if image_href is ordinary file name, False otherwise.
    """
    return not (service_utils.is_glance_image(image_href)
                or urlparse.urlparse(image_href).scheme.lower() in
                image_service.protocol_mapping)


def _parse_config_option():
    """Parse config file options.

    This method checks config file options validity.

    :raises: InvalidParameterValue, if config option has invalid value.
    """
    error_msgs = []
    if not os.path.isdir(CONF.irmc.remote_image_share_root):
        error_msgs.append(
            _("Value '%s' for remote_image_share_root isn't a directory "
              "or doesn't exist.") %
            CONF.irmc.remote_image_share_root)
    if error_msgs:
        msg = (_("The following errors were encountered while parsing "
                 "config file:%s") % error_msgs)
        raise exception.InvalidParameterValue(msg)


def _parse_driver_info(node, mode='deploy'):
    """Gets the driver specific Node deployment info.

    This method validates whether the 'driver_info' property of the
    supplied node contains the required or optional information properly
    for this driver to deploy images to the node.

    :param node: a target node of the deployment
    :param mode: Label indicating a deploy or rescue operation being
                 carried out on the node. Supported values are
                 'deploy' and 'rescue'. Defaults to 'deploy'.
    :returns: the driver_info values of the node.
    :raises: MissingParameterValue, if any of the required parameters are
        missing.
    :raises: InvalidParameterValue, if any of the parameters have invalid
        value.
    """
    deploy_info = {}

    image_iso = driver_utils.get_agent_iso(node, mode,
                                           deprecated_prefix='irmc')
    deploy_info[f'{mode}_iso'] = image_iso

    error_msg = (_("Error validating iRMC virtual media for %s. Some "
                   "parameters were missing in node's driver_info") % mode)
    deploy_utils.check_for_missing_params(deploy_info, error_msg)

    if _is_image_href_ordinary_file_name(image_iso):
        image_iso_file = os.path.join(CONF.irmc.remote_image_share_root,
                                      image_iso)
        if not os.path.isfile(image_iso_file):
            msg = (_("%(mode)s ISO file, %(iso_file)s, "
                     "not found for node: %(node)s.") %
                   {'mode': mode.capitalize(),
                    'iso_file': image_iso_file,
                    'node': node.uuid})
            raise exception.InvalidParameterValue(msg)

    kernel_params = driver_utils.get_kernel_append_params(
        node, default=CONF.irmc.kernel_append_params)
    if kernel_params is None:
        LOG.warning('Relying on [pxe]kernel_append_params in the iRMC '
                    'hardware type is deprecated, please set '
                    '[irmc]kernel_append_params')
        kernel_params = CONF.pxe.kernel_append_params
    deploy_info['kernel_append_params'] = kernel_params

    return deploy_info


def _parse_instance_info(node):
    """Gets the instance specific Node deployment info.

    This method validates whether the 'instance_info' property of the
    supplied node contains the required or optional information properly
    for this driver to deploy images to the node.

    :param node: a target node of the deployment
    :returns:  the instance_info values of the node.
    :raises: InvalidParameterValue, if any of the parameters have invalid
        value.
    """
    deploy_info = {}

    boot_iso = driver_utils.get_field(node, 'boot_iso',
                                      deprecated_prefix='irmc',
                                      collection='instance_info')
    if boot_iso:
        deploy_info['boot_iso'] = boot_iso

        if _is_image_href_ordinary_file_name(boot_iso):
            boot_iso = os.path.join(CONF.irmc.remote_image_share_root,
                                    boot_iso)

            if not os.path.isfile(boot_iso):
                msg = (_("Boot ISO file, %(boot_iso)s, "
                         "not found for node: %(node)s.") %
                       {'boot_iso': boot_iso, 'node': node.uuid})
                raise exception.InvalidParameterValue(msg)

    return deploy_info


def _parse_deploy_info(node):
    """Gets the instance and driver specific Node deployment info.

    This method validates whether the 'instance_info' and 'driver_info'
    property of the supplied node contains the required information for
    this driver to deploy images to the node.

    :param node: a target node of the deployment
    :returns: a dict with the instance_info and driver_info values.
    :raises: MissingParameterValue, if any of the required parameters are
        missing.
    :raises: InvalidParameterValue, if any of the parameters have invalid
        value.
    """
    deploy_info = {}
    deploy_info.update(deploy_utils.get_image_instance_info(node))
    deploy_info.update(_parse_driver_info(node))
    deploy_info.update(_parse_instance_info(node))

    return deploy_info


def _setup_vmedia(task, mode, ramdisk_options):
    """Attaches virtual media and sets it as boot device.

    This method attaches the deploy or rescue ISO as virtual media, prepares
    the arguments for ramdisk in virtual media floppy.

    :param task: a TaskManager instance containing the node to act on.
    :param mode: Label indicating a deploy or rescue operation being
                 carried out on the node. Supported values are
                 'deploy' and 'rescue'.
    :param ramdisk_options: the options to be passed to the ramdisk in virtual
        media floppy.
    :raises: ImageRefValidationFailed if no image service can handle specified
       href.
    :raises: ImageCreationFailed, if it failed while creating the floppy image.
    :raises: IRMCOperationError, if some operation on iRMC failed.
    :raises: InvalidParameterValue if the validation of the
        PowerInterface or ManagementInterface fails.
    """
    iso = driver_utils.get_agent_iso(task.node, mode, deprecated_prefix='irmc')

    if _is_image_href_ordinary_file_name(iso):
        iso_file = iso
    else:
        iso_file = _get_iso_name(task.node, label=mode)
        iso_fullpathname = os.path.join(
            CONF.irmc.remote_image_share_root, iso_file)
        images.fetch(task.context, iso, iso_fullpathname)

    _setup_vmedia_for_boot(task, iso_file, ramdisk_options)
    manager_utils.node_set_boot_device(task, boot_devices.CDROM)


def _get_iso_name(node, label):
    """Returns the ISO file name for a given node.

    :param node: the node for which ISO file name is to be provided.
    :param label: a string used as a base name for the ISO file.
    """
    return "%s-%s.iso" % (label, node.uuid)


def _prepare_boot_iso(task, root_uuid):
    """Prepare a boot ISO to boot the node.

    :param task: a TaskManager instance containing the node to act on.
    :param root_uuid: the uuid of the root partition.
    :raises: MissingParameterValue, if any of the required parameters are
        missing.
    :raises: InvalidParameterValue, if any of the parameters have invalid
        value.
    :raises: ImageCreationFailed, if creating boot ISO
       for BIOS boot_mode failed.
    """
    deploy_info = _parse_deploy_info(task.node)

    # fetch boot iso
    if deploy_info.get('boot_iso'):
        boot_iso_href = deploy_info['boot_iso']
        if _is_image_href_ordinary_file_name(boot_iso_href):
            task.node.set_driver_internal_info('boot_iso', boot_iso_href)
        else:
            boot_iso_filename = _get_iso_name(task.node, label='boot')
            boot_iso_fullpathname = os.path.join(
                CONF.irmc.remote_image_share_root, boot_iso_filename)
            images.fetch(task.context, boot_iso_href, boot_iso_fullpathname)

            task.node.set_driver_internal_info('boot_iso',
                                               boot_iso_filename)

    # create boot iso
    else:
        image_href = deploy_info['image_source']
        image_props = ['kernel_id', 'ramdisk_id']
        image_properties = images.get_image_properties(
            task.context, image_href, image_props)
        kernel_href = (task.node.instance_info.get('kernel')
                       or image_properties['kernel_id'])
        ramdisk_href = (task.node.instance_info.get('ramdisk')
                        or image_properties['ramdisk_id'])

        deploy_iso_href = deploy_info['deploy_iso']
        boot_mode = boot_mode_utils.get_boot_mode(task.node)
        kernel_params = deploy_info['kernel_append_params']

        boot_iso_filename = _get_iso_name(task.node, label='boot')
        boot_iso_fullpathname = os.path.join(
            CONF.irmc.remote_image_share_root, boot_iso_filename)

        images.create_boot_iso(task.context, boot_iso_fullpathname,
                               kernel_href, ramdisk_href,
                               deploy_iso_href=deploy_iso_href,
                               root_uuid=root_uuid,
                               kernel_params=kernel_params,
                               boot_mode=boot_mode)

        task.node.set_driver_internal_info('boot_iso',
                                           boot_iso_filename)

    # save driver_internal_info['boot_iso']
    task.node.save()


def _get_floppy_image_name(node):
    """Returns the floppy image name for a given node.

    :param node: the node for which image name is to be provided.
    """
    return "image-%s.img" % node.uuid


def _prepare_floppy_image(task, params):
    """Prepares the floppy image for passing the parameters.

    This method prepares a temporary vfat filesystem image, which
    contains the parameters to be passed to the ramdisk.
    Then it uploads the file NFS or CIFS server.

    :param task: a TaskManager instance containing the node to act on.
    :param params: a dictionary containing 'parameter name'->'value' mapping
        to be passed to the deploy ramdisk via the floppy image.
    :returns: floppy image filename
    :raises: ImageCreationFailed, if it failed while creating the floppy image.
    :raises: IRMCOperationError, if copying floppy image file failed.
    """
    floppy_filename = _get_floppy_image_name(task.node)
    floppy_fullpathname = os.path.join(
        CONF.irmc.remote_image_share_root, floppy_filename)

    with tempfile.NamedTemporaryFile() as vfat_image_tmpfile_obj:
        images.create_vfat_image(vfat_image_tmpfile_obj.name,
                                 parameters=params)
        try:
            shutil.copyfile(vfat_image_tmpfile_obj.name,
                            floppy_fullpathname)
        except IOError as e:
            operation = _("Copying floppy image file")
            raise exception.IRMCOperationError(
                operation=operation, error=e)

    return floppy_filename


def attach_boot_iso_if_needed(task):
    """Attaches boot ISO for a deployed node if it exists.

    This method checks the instance info of the bare metal node for a
    boot ISO. If the instance info has a value of key 'boot_iso',
    it indicates ramdisk deploy. Therefore it attaches the boot ISO on the bare
    metal node and then sets the node to boot from virtual media cdrom.

    :param task: a TaskManager instance containing the node to act on.
    :raises: IRMCOperationError if attaching virtual media failed.
    :raises: InvalidParameterValue if the validation of the
        ManagementInterface fails.
    """
    d_info = task.node.driver_internal_info
    node_state = task.node.provision_state

    # Internal field, no deprecation
    boot_iso = d_info.get('boot_iso') or d_info.get('irmc_boot_iso')
    if boot_iso and node_state == states.ACTIVE:
        _setup_vmedia_for_boot(task, boot_iso)
        manager_utils.node_set_boot_device(task, boot_devices.CDROM)


def _setup_vmedia_for_boot(task, bootable_iso_filename, parameters=None):
    """Sets up the node to boot from the boot ISO image.

    This method attaches a boot_iso on the node and passes
    the required parameters to it via a virtual floppy image.

    :param task: a TaskManager instance containing the node to act on.
    :param bootable_iso_filename: a bootable ISO image to attach to.
        The iso file should be present in NFS/CIFS server.
    :param parameters: the parameters to pass in a virtual floppy image
        in a dictionary.  This is optional.
    :raises: ImageCreationFailed, if it failed while creating a floppy image.
    :raises: IRMCOperationError, if attaching a virtual media failed.
    """
    LOG.info("Setting up node %s to boot from virtual media",
             task.node.uuid)

    _detach_virtual_cd(task.node)
    _detach_virtual_fd(task.node)

    if parameters:
        floppy_image_filename = _prepare_floppy_image(task, parameters)
        _attach_virtual_fd(task.node, floppy_image_filename)

    _attach_virtual_cd(task.node, bootable_iso_filename)


def _cleanup_vmedia_boot(task):
    """Cleans a node after a virtual media boot.

    This method cleans up a node after a virtual media boot.
    It deletes floppy and cdrom images if they exist in NFS/CIFS server.
    It also ejects both the virtual media cdrom and the virtual media floppy.

    :param task: a TaskManager instance containing the node to act on.
    :raises: IRMCOperationError if ejecting virtual media failed.
    """
    LOG.debug("Cleaning up node %s after virtual media boot", task.node.uuid)

    node = task.node
    _detach_virtual_cd(node)
    _detach_virtual_fd(node)

    _remove_share_file(_get_floppy_image_name(node))
    _remove_share_file(_get_iso_name(node, label='deploy'))
    _remove_share_file(_get_iso_name(node, label='rescue'))


def _remove_share_file(share_filename):
    """Remove given file from the share file system.

    :param share_filename: a file name to be removed.
    """
    share_fullpathname = os.path.join(
        CONF.irmc.remote_image_share_root, share_filename)
    utils.unlink_without_raise(share_fullpathname)


def _attach_virtual_cd(node, bootable_iso_filename):
    """Attaches the given url as virtual media on the node.

    :param node: an ironic node object.
    :param bootable_iso_filename: a bootable ISO image to attach to.
        The iso file should be present in NFS/CIFS server.
    :raises: IRMCOperationError if attaching virtual media failed.
    """
    try:
        irmc_client = irmc_common.get_irmc_client(node)

        cd_set_params = scci.get_virtual_cd_set_params_cmd(
            CONF.irmc.remote_image_server,
            CONF.irmc.remote_image_user_domain,
            scci.get_share_type(CONF.irmc.remote_image_share_type),
            CONF.irmc.remote_image_share_name,
            bootable_iso_filename,
            CONF.irmc.remote_image_user_name,
            CONF.irmc.remote_image_user_password)

        irmc_client(cd_set_params, do_async=False)
        irmc_client(scci.MOUNT_CD, do_async=False)

    except scci.SCCIClientError as irmc_exception:
        LOG.exception("Error while inserting virtual cdrom "
                      "into node %(uuid)s. Error: %(error)s",
                      {'uuid': node.uuid, 'error': irmc_exception})
        operation = _("Inserting virtual cdrom")
        raise exception.IRMCOperationError(operation=operation,
                                           error=irmc_exception)

    LOG.info("Attached virtual cdrom successfully"
             " for node %s", node.uuid)


def _detach_virtual_cd(node):
    """Detaches virtual cdrom on the node.

    :param node: an ironic node object.
    :raises: IRMCOperationError if eject virtual cdrom failed.
    """
    try:
        irmc_client = irmc_common.get_irmc_client(node)

        irmc_client(scci.UNMOUNT_CD)

    except scci.SCCIClientError as irmc_exception:
        LOG.exception("Error while ejecting virtual cdrom "
                      "from node %(uuid)s. Error: %(error)s",
                      {'uuid': node.uuid, 'error': irmc_exception})
        operation = _("Ejecting virtual cdrom")
        raise exception.IRMCOperationError(operation=operation,
                                           error=irmc_exception)

    LOG.info("Detached virtual cdrom successfully"
             " for node %s", node.uuid)


def _attach_virtual_fd(node, floppy_image_filename):
    """Attaches virtual floppy on the node.

    :param node: an ironic node object.
    :raises: IRMCOperationError if insert virtual floppy failed.
    """
    try:
        irmc_client = irmc_common.get_irmc_client(node)

        fd_set_params = scci.get_virtual_fd_set_params_cmd(
            CONF.irmc.remote_image_server,
            CONF.irmc.remote_image_user_domain,
            scci.get_share_type(CONF.irmc.remote_image_share_type),
            CONF.irmc.remote_image_share_name,
            floppy_image_filename,
            CONF.irmc.remote_image_user_name,
            CONF.irmc.remote_image_user_password)

        irmc_client(fd_set_params, do_async=False)
        irmc_client(scci.MOUNT_FD, do_async=False)

    except scci.SCCIClientError as irmc_exception:
        LOG.exception("Error while inserting virtual floppy "
                      "into node %(uuid)s. Error: %(error)s",
                      {'uuid': node.uuid, 'error': irmc_exception})
        operation = _("Inserting virtual floppy")
        raise exception.IRMCOperationError(operation=operation,
                                           error=irmc_exception)

    LOG.info("Attached virtual floppy successfully"
             " for node %s", node.uuid)


def _detach_virtual_fd(node):
    """Detaches virtual media floppy on the node.

    :param node: an ironic node object.
    :raises: IRMCOperationError if eject virtual media floppy failed.
    """
    try:
        irmc_client = irmc_common.get_irmc_client(node)

        irmc_client(scci.UNMOUNT_FD)

    except scci.SCCIClientError as irmc_exception:
        LOG.exception("Error while ejecting virtual floppy "
                      "from node %(uuid)s. Error: %(error)s",
                      {'uuid': node.uuid, 'error': irmc_exception})
        operation = _("Ejecting virtual floppy")
        raise exception.IRMCOperationError(operation=operation,
                                           error=irmc_exception)

    LOG.info("Detached virtual floppy successfully"
             " for node %s", node.uuid)


def check_share_fs_mounted():
    """Check if Share File System (NFS or CIFS) is mounted.

    :raises: InvalidParameterValue, if config option has invalid value.
    :raises: IRMCSharedFileSystemNotMounted, if shared file system is
        not mounted.
    """
    _parse_config_option()
    if not os.path.ismount(CONF.irmc.remote_image_share_root):
        raise exception.IRMCSharedFileSystemNotMounted(
            share=CONF.irmc.remote_image_share_root)


class IRMCVolumeBootMixIn(object):
    """Mix-in class for volume boot configuration to iRMC

    iRMC has a feature to set up remote boot to a server. This feature can be
    used by VIOM (Virtual I/O Manager) library of SCCI client.
    """

    def _validate_volume_boot(self, task):
        """Validate information for volume boot with this interface.

        This interface requires physical information of connectors to
        configure remote boot to iRMC. Physical information of LAN ports
        is also required since VIOM feature manages all adapters.

        :param task: a TaskManager instance containing the node to act on.
        :raises: InvalidParameterValue: If invalid value is set to resources.
        :raises: MissingParameterValue: If some value is not set to resources.
        """

        if not deploy_utils.get_remote_boot_volume(task):
            # No boot volume. Nothing to validate.
            return

        irmc_common.parse_driver_info(task.node)

        for port in task.ports:
            self._validate_lan_port(task.node, port)

        for vt in task.volume_targets:
            if vt.volume_type == 'iscsi':
                self._validate_iscsi_connectors(task)
            elif vt.volume_type == 'fibre_channel':
                self._validate_fc_connectors(task)
            # Unknown volume type is filtered in storage interface validation.

    def _get_connector_physical_id(self, task, types):
        """Get physical ID of volume connector.

        A physical ID of volume connector required by iRMC is registered in
        "irmc_pci_physical_ids" of a Node's driver_info as a pair of resource
        UUID and its physical ID. This method gets this ID from the parameter.

        :param task: a TaskManager instance containing the node to act on.
        :param types: a list of types of volume connectors required for the
            target volume. One of connectors must have a physical ID.
        :raises InvalidParameterValue if a physical ID is invalid.
        :returns: A physical ID of a volume connector, or None if not set.
        """
        for vc in task.volume_connectors:
            if vc.type not in types:
                continue
            pid = task.node.driver_info['irmc_pci_physical_ids'].get(vc.uuid)
            if not pid:
                continue
            try:
                viom.validate_physical_port_id(pid)
            except scci.SCCIInvalidInputError as e:
                raise exception.InvalidParameterValue(
                    _('Physical port information of volume connector '
                      '%(connector)s is invalid: %(error)s') %
                    {'connector': vc.uuid, 'error': e})
            return pid
        return None

    def _validate_iscsi_connectors(self, task):
        """Validate if volume connectors are properly registered for iSCSI.

        For connecting a node to an iSCSI volume, volume connectors containing
        an IQN and an IP address are necessary. One of connectors must have
        a physical ID of the PCI card. Network size of a storage network is
        also required by iRMC. which should be registered in the node's
        driver_info.

        :param task: a TaskManager instance containing the node to act on.
        :raises: InvalidParameterValue if a volume connector with a required
            type is not registered.
        :raises: InvalidParameterValue if a physical ID is not registered in
            any volume connectors.
        :raises: InvalidParameterValue if a physical ID is invalid.
        """
        vc_dict = self._get_volume_connectors_by_type(task)
        node = task.node
        missing_types = []
        for vc_type in ('iqn', 'ip'):
            vc = vc_dict.get(vc_type)
            if not vc:
                missing_types.append(vc_type)

        if missing_types:
            raise exception.MissingParameterValue(
                _('Failed to validate for node %(node)s because of missing '
                  'volume connector(s) with type(s) %(types)s') %
                {'node': node.uuid,
                 'types': ', '.join(missing_types)})

        if not self._get_connector_physical_id(task, ['iqn', 'ip']):
            raise exception.MissingParameterValue(
                _('Failed to validate for node %(node)s because of missing '
                  'physical port information for iSCSI connector. This '
                  'information must be set in "pci_physical_ids" parameter of '
                  'node\'s driver_info as <connector uuid>:<physical id>.') %
                {'node': node.uuid})
        self._get_network_size(node)

    def _validate_fc_connectors(self, task):
        """Validate if volume connectors are properly registered for FC.

        For connecting a node to a FC volume, one of connectors representing
        wwnn and wwpn must have a physical ID of the PCI card.

        :param task: a TaskManager instance containing the node to act on.
        :raises: InvalidParameterValue if a physical ID is not registered in
            any volume connectors.
        :raises: InvalidParameterValue if a physical ID is invalid.
        """
        node = task.node
        if not self._get_connector_physical_id(task, ['wwnn', 'wwpn']):
            raise exception.MissingParameterValue(
                _('Failed to validate for node %(node)s because of missing '
                  'physical port information for FC connector. This '
                  'information must be set in "pci_physical_ids" parameter of '
                  'node\'s driver_info as <connector uuid>:<physical id>.') %
                {'node': node.uuid})

    def _validate_lan_port(self, node, port):
        """Validate ports for VIOM configuration.

        Physical information of LAN ports must be registered to VIOM
        configuration to activate them under VIOM management. The information
        has to be set to "irmc_pci_physical_id" parameter in a nodes
        driver_info.

        :param node: an ironic node object
        :param port: a port to be validated
        :raises: MissingParameterValue if a physical ID of the port is not set.
        :raises: InvalidParameterValue if a physical ID is invalid.
        """
        physical_id = node.driver_info['irmc_pci_physical_ids'].get(port.uuid)
        if not physical_id:
            raise exception.MissingParameterValue(
                _('Failed to validate for node %(node)s because of '
                  'missing physical port information of port %(port)s. '
                  'This information should be contained in '
                  '"pci_physical_ids" parameter of node\'s driver_info.') %
                {'node': node.uuid,
                 'port': port.uuid})
        try:
            viom.validate_physical_port_id(physical_id)
        except scci.SCCIInvalidInputError as e:
            raise exception.InvalidParameterValue(
                _('Failed to validate for node %(node)s because '
                  'the physical port ID for port %(port)s in node\'s'
                  ' driver_info is invalid: %(reason)s') %
                {'node': node.uuid,
                 'port': port.uuid,
                 'reason': e})

    def _get_network_size(self, node):
        """Get network size of a storage network.

        The network size of iSCSI network is required by iRMC for connecting
        a node to an iSCSI volume. This network size is set to node's
        driver_info as "irmc_storage_network_size" parameter in the form of
        positive integer.

        :param node: an ironic node object.
        :raises: MissingParameterValue if the network size parameter is not
            set.
        :raises: InvalidParameterValue the network size is invalid.
        """
        network_size = node.driver_info.get('irmc_storage_network_size')
        if network_size is None:
            raise exception.MissingParameterValue(
                _('Failed to validate for node %(node)s because of '
                  'missing "irmc_storage_network_size" parameter in the '
                  'node\'s driver_info. This should be a positive integer '
                  'smaller than 32.') %
                {'node': node.uuid})
        try:
            network_size = int(network_size)
        except (ValueError, TypeError):
            raise exception.InvalidParameterValue(
                _('Failed to validate for node %(node)s because '
                  '"irmc_storage_network_size" parameter in the node\'s '
                  'driver_info is invalid. This should be a '
                  'positive integer smaller than 32.') %
                {'node': node.uuid})

        if network_size not in range(1, 32):
            raise exception.InvalidParameterValue(
                _('Failed to validate for node %(node)s because '
                  '"irmc_storage_network_size" parameter in the node\'s '
                  'driver_info is invalid. This should be a '
                  'positive integer smaller than 32.') %
                {'node': node.uuid})

        return network_size

    def _get_volume_connectors_by_type(self, task):
        """Create a dictionary of volume connectors by types.

        :param task: a TaskManager.
        :returns: a volume connector dictionary whose key is a connector type.
        """
        connectors = {}
        for vc in task.volume_connectors:
            if vc.type in ('ip', 'iqn', 'wwnn', 'wwpn'):
                connectors[vc.type] = vc
            else:
                LOG.warning('Node %(node)s has a volume_connector (%(uuid)s) '
                            'defined with an unsupported type: %(type)s.',
                            {'node': task.node.uuid,
                             'uuid': vc.uuid,
                             'type': vc.type})
        return connectors

    def _register_lan_ports(self, viom_conf, task):
        """Register ports to VIOM configuration.

        LAN ports information must be registered for VIOM configuration to
        activate them under VIOM management.

        :param viom_conf: a configurator for iRMC
        :param task: a TaskManager instance containing the node to act on.
        """
        for port in task.ports:
            viom_conf.set_lan_port(
                task.node.driver_info['irmc_pci_physical_ids'].get(port.uuid))

    def _configure_boot_from_volume(self, task):
        """Set information for booting from a remote volume to iRMC.

        :param task: a TaskManager instance containing the node to act on.
        :raises: IRMCOperationError if iRMC operation failed
        """

        irmc_info = irmc_common.parse_driver_info(task.node)
        viom_conf = viom.VIOMConfiguration(irmc_info,
                                           identification=task.node.uuid)

        self._register_lan_ports(viom_conf, task)

        for vt in task.volume_targets:
            if vt.volume_type == 'iscsi':
                self._set_iscsi_target(task, viom_conf, vt)
            elif vt.volume_type == 'fibre_channel':
                self._set_fc_target(task, viom_conf, vt)

        try:
            LOG.debug('Set VIOM configuration for node %(node)s: %(table)s',
                      {'node': task.node.uuid,
                       'table': viom_conf.dump_json()})
            viom_conf.apply()
        except scci.SCCIError as e:
            LOG.error('iRMC failed to set VIOM configuration for node '
                      '%(node)s: %(error)s',
                      {'node': task.node.uuid,
                       'error': e})
            raise exception.IRMCOperationError(
                operation='Configure VIOM', error=e)

    def _set_iscsi_target(self, task, viom_conf, target):
        """Set information for iSCSI boot to VIOM configuration."""
        connectors = self._get_volume_connectors_by_type(task)
        target_host, target_port = netutils.parse_host_port(
            target.properties['target_portal'])

        if target.properties.get('auth_method') == 'CHAP':
            chap_user = target.properties.get('auth_username')
            chap_secret = target.properties.get('auth_password')
        else:
            chap_user = None
            chap_secret = None

        viom_conf.set_iscsi_volume(
            self._get_connector_physical_id(task, ['iqn', 'ip']),
            connectors['iqn'].connector_id,
            initiator_ip=connectors['ip'].connector_id,
            initiator_netmask=self._get_network_size(task.node),
            target_iqn=target.properties['target_iqn'],
            target_ip=target_host,
            target_port=target_port,
            target_lun=target.properties.get('target_lun'),
            # Boot priority starts from 1 in the library.
            boot_prio=target.boot_index + 1,
            chap_user=chap_user,
            chap_secret=chap_secret)

    def _set_fc_target(self, task, viom_conf, target):
        """Set information for FC boot to VIOM configuration."""
        wwn = target.properties['target_wwn']
        if isinstance(wwn, list):
            wwn = wwn[0]
        viom_conf.set_fc_volume(
            self._get_connector_physical_id(task, ['wwnn', 'wwpn']),
            wwn,
            target.properties['target_lun'],
            # Boot priority starts from 1 in the library.
            boot_prio=target.boot_index + 1)

    def _cleanup_boot_from_volume(self, task, reboot=False):
        """Clear remote boot configuration.

        :param task: a task from TaskManager.
        :param reboot: True if reboot node soon
        :raises: IRMCOperationError if iRMC operation failed
        """
        irmc_info = irmc_common.parse_driver_info(task.node)
        try:
            viom_conf = viom.VIOMConfiguration(irmc_info, task.node.uuid)
            viom_conf.terminate(reboot=reboot)
        except scci.SCCIError as e:
            LOG.error('iRMC failed to terminate VIOM configuration from '
                      'node %(node)s: %(error)s', {'node': task.node.uuid,
                                                   'error': e})
            raise exception.IRMCOperationError(operation='Terminate VIOM',
                                               error=e)


class IRMCVirtualMediaBoot(base.BootInterface, IRMCVolumeBootMixIn):
    """iRMC Virtual Media boot-related actions."""

    capabilities = ['iscsi_volume_boot', 'fibre_channel_volume_boot']

    def __init__(self):
        """Constructor of IRMCVirtualMediaBoot.

        :raises: IRMCSharedFileSystemNotMounted, if shared file system is
            not mounted.
        :raises: InvalidParameterValue, if config option has invalid value.
        """
        check_share_fs_mounted()
        super(IRMCVirtualMediaBoot, self).__init__()

    def get_properties(self):
        # TODO(tiendc): COMMON_PROPERTIES should also include rescue
        # related properties (RESCUE_PROPERTIES). We can add them in Rocky,
        # when classic drivers get removed.
        return COMMON_PROPERTIES

    @METRICS.timer('IRMCVirtualMediaBoot.validate')
    def validate(self, task):
        """Validate the deployment information for the task's node.

        :param task: a TaskManager instance containing the node to act on.
        :raises: InvalidParameterValue, if config option has invalid value.
        :raises: IRMCSharedFileSystemNotMounted, if shared file system is
            not mounted.
        :raises: InvalidParameterValue, if some information is invalid.
        :raises: MissingParameterValue if 'kernel_id' and 'ramdisk_id' are
            missing in the Glance image, or if 'kernel' and 'ramdisk' are
            missing in the Non Glance image.
        """
        check_share_fs_mounted()

        self._validate_volume_boot(task)
        if not task.driver.storage.should_write_image(task):
            LOG.debug('Node %(node)s skips image validation because of '
                      'booting from a remote volume.',
                      {'node': task.node.uuid})
            return

        d_info = _parse_deploy_info(task.node)
        deploy_utils.validate_image_properties(task, d_info)

    @METRICS.timer('IRMCVirtualMediaBoot.prepare_ramdisk')
    def prepare_ramdisk(self, task, ramdisk_params):
        """Prepares the deploy or rescue ramdisk using virtual media.

        Prepares the options for the deploy or rescue ramdisk, sets the node
        to boot from virtual media cdrom.

        :param task: a TaskManager instance containing the node to act on.
        :param ramdisk_params: the options to be passed to the ramdisk.
        :raises: ImageRefValidationFailed if no image service can handle
                 specified href.
        :raises: ImageCreationFailed, if it failed while creating the floppy
                 image.
        :raises: InvalidParameterValue if the validation of the
                 PowerInterface or ManagementInterface fails.
        :raises: IRMCOperationError, if some operation on iRMC fails.
        """

        if not driver_utils.need_prepare_ramdisk(task.node):
            return

        # NOTE(tiendc): Before deploying, we need to backup BIOS config
        # as the data will be used later when cleaning.
        if task.node.provision_state == states.DEPLOYING:
            irmc_management.backup_bios_config(task)

            if not task.driver.storage.should_write_image(task):
                LOG.debug('Node %(node)s skips ramdisk preparation because of '
                          'booting from a remote volume.',
                          {'node': task.node.uuid})
                return

        # NOTE(TheJulia): Since we're deploying, cleaning, or rescuing,
        # with virtual media boot, we should generate a token!
        manager_utils.add_secret_token(task.node, pregenerated=True)
        ramdisk_params['ipa-agent-token'] = \
            task.node.driver_internal_info['agent_secret_token']
        task.node.save()

        deploy_nic_mac = deploy_utils.get_single_nic_with_vif_port_id(task)
        if deploy_nic_mac is not None:
            ramdisk_params['BOOTIF'] = deploy_nic_mac

        if task.node.provision_state == states.RESCUING:
            mode = 'rescue'
        else:
            mode = 'deploy'

        _setup_vmedia(task, mode, ramdisk_params)

    @METRICS.timer('IRMCVirtualMediaBoot.clean_up_ramdisk')
    def clean_up_ramdisk(self, task):
        """Cleans up the boot of ironic ramdisk.

        This method cleans up the environment that was setup for booting the
        deploy or rescue ramdisk.

        :param task: a task from TaskManager.
        :returns: None
        :raises: IRMCOperationError if iRMC operation failed.
        """
        _cleanup_vmedia_boot(task)

    @METRICS.timer('IRMCVirtualMediaBoot.prepare_instance')
    def prepare_instance(self, task):
        """Prepares the boot of instance.

        This method prepares the boot of the instance after reading
        relevant information from the node's database.

        :param task: a task from TaskManager.
        :returns: None
        """
        if task.node.driver_internal_info.get('boot_from_volume'):
            LOG.debug('Node %(node)s is configured for booting from a remote '
                      'volume.',
                      {'node': task.node.uuid})
            self._configure_boot_from_volume(task)
            return

        _cleanup_vmedia_boot(task)

        node = task.node
        iwdi = node.driver_internal_info.get('is_whole_disk_image')
        if deploy_utils.get_boot_option(node) == "local" or iwdi:
            manager_utils.node_set_boot_device(task, boot_devices.DISK,
                                               persistent=True)
        else:
            root_uuid_or_disk_id = node.driver_internal_info[
                'root_uuid_or_disk_id']
            self._configure_vmedia_boot(task, root_uuid_or_disk_id)

        # Enable secure boot, if being requested
        boot_mode_utils.configure_secure_boot_if_needed(task)

    @METRICS.timer('IRMCVirtualMediaBoot.clean_up_instance')
    def clean_up_instance(self, task):
        """Cleans up the boot of instance.

        This method cleans up the environment that was setup for booting
        the instance.

        :param task: a task from TaskManager.
        :returns: None
        :raises: IRMCOperationError if iRMC operation failed.
        """
        if task.node.driver_internal_info.get('boot_from_volume'):
            self._cleanup_boot_from_volume(task)
            return

        # Disable secure boot, if enabled secure boot
        boot_mode_utils.deconfigure_secure_boot_if_needed(task)

        _remove_share_file(_get_iso_name(task.node, label='boot'))
        task.node.del_driver_internal_info('boot_iso')
        task.node.del_driver_internal_info('irmc_boot_iso')

        task.node.save()
        _cleanup_vmedia_boot(task)

    def _configure_vmedia_boot(self, task, root_uuid_or_disk_id):
        """Configure vmedia boot for the node."""
        node = task.node
        _prepare_boot_iso(task, root_uuid_or_disk_id)
        _setup_vmedia_for_boot(
            task, node.driver_internal_info['boot_iso'])
        manager_utils.node_set_boot_device(task, boot_devices.CDROM,
                                           persistent=True)

    @METRICS.timer('IRMCVirtualMediaBoot.validate_rescue')
    def validate_rescue(self, task):
        """Validate that the node has required properties for rescue.

        :param task: a TaskManager instance with the node being checked
        :raises: MissingParameterValue if node is missing one or more required
            parameters
        :raises: InvalidParameterValue, if any of the parameters have invalid
            value.
        """
        _parse_driver_info(task.node, mode='rescue')


class IRMCPXEBoot(pxe.PXEBoot):
    """iRMC PXE boot."""

    @METRICS.timer('IRMCPXEBoot.prepare_ramdisk')
    def prepare_ramdisk(self, task, ramdisk_params):
        """Prepares the boot of Ironic ramdisk using PXE.

        This method prepares the boot of the deploy kernel/ramdisk after
        reading relevant information from the node's driver_info and
        instance_info.

        :param task: a task from TaskManager.
        :param ramdisk_params: the parameters to be passed to the ramdisk.
            pxe driver passes these parameters as kernel command-line
            arguments.
        :returns: None
        :raises: MissingParameterValue, if some information is missing in
            node's driver_info or instance_info.
        :raises: InvalidParameterValue, if some information provided is
            invalid.
        :raises: IronicException, if some power or set boot device
            operation failed on the node.
        """
        # NOTE(tiendc): Before deploying, we need to backup BIOS config
        # as the data will be used later when cleaning.
        if task.node.provision_state == states.DEPLOYING:
            irmc_management.backup_bios_config(task)

        super(IRMCPXEBoot, self).prepare_ramdisk(task, ramdisk_params)



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_concurrency import lockutils
from oslo_log import log as logging
import stevedore

from ironic.common import exception
from ironic.conf import CONF

EM_SEMAPHORE = 'console_container_provider'

LOG = logging.getLogger(__name__)


class ConsoleContainerFactory(object):

    _provider = None

    def __init__(self, **kwargs):
        if not ConsoleContainerFactory._provider:
            ConsoleContainerFactory._set_provider(**kwargs)

    # Use lockutils to avoid a potential race in eventlet
    # that might try to create two factories.
    @classmethod
    @lockutils.synchronized(EM_SEMAPHORE)
    def _set_provider(cls, **kwargs):
        """Initialize the provider

        :raises: ConsoleContainerError if the provider cannot be loaded.
        """

        # In case multiple greenthreads queue up on
        # this lock before _provider is initialized,
        # prevent creation of multiple DriverManager.
        if cls._provider:
            return

        provider_name = CONF.vnc.container_provider
        try:
            _extension_manager = stevedore.driver.DriverManager(
                'ironic.console.container',
                provider_name,
                invoke_kwds=kwargs,
                invoke_on_load=True)
        except Exception as e:
            LOG.exception('Could not create console container provider')
            raise exception.ConsoleContainerError(
                provider=provider_name, reason=e
            )

        cls._provider = _extension_manager.driver

    @property
    def provider(self):
        return self._provider



#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os
import tempfile
from unittest import mock

from oslo_concurrency import processutils
from oslo_config import cfg

from ironic.common import console_factory
from ironic.common import exception
from ironic.common import utils
from ironic.console.container import fake
from ironic.tests import base

CONF = cfg.CONF


def _reset_provider(provider_name):
    CONF.set_override('container_provider', provider_name, 'vnc')
    console_factory.ConsoleContainerFactory._provider = None


class TestConsoleContainerFactory(base.TestCase):

    def setUp(self):
        super(TestConsoleContainerFactory, self).setUp()
        _reset_provider('fake')

    def test_factory(self):
        provider = console_factory.ConsoleContainerFactory().provider

        self.assertIsInstance(provider, fake.FakeConsoleContainer)

        provider2 = console_factory.ConsoleContainerFactory().provider
        self.assertEqual(provider, provider2)


class TestSystemdConsoleContainer(base.TestCase):

    def setUp(self):
        super(TestSystemdConsoleContainer, self).setUp()
        _reset_provider('systemd')
        self.addCleanup(_reset_provider, 'fake')
        self.tempdir = tempfile.mkdtemp()
        self.addCleanup(lambda: utils.rmtree_without_raise(self.tempdir))
        os.environ['XDG_RUNTIME_DIR'] = self.tempdir
        with mock.patch.object(utils, 'execute', autospec=True) as mock_exec:
            self.provider = console_factory.ConsoleContainerFactory().provider
            mock_exec.assert_has_calls([
                mock.call('systemctl', '--version'),
                mock.call('podman', '--version'),
            ])
        # Override unit directory with tempdir
        self.provider._init_unit_dir(self.tempdir)

    def test__container_path(self):
        self.assertEqual(
            f'{self.tempdir}/ironic-console-1234.container',
            self.provider._container_path('1234'))

    def test__unit_name(self):
        self.assertEqual(
            'ironic-console-1234.service',
            self.provider._unit_name('1234')
        )

    def test__container_name(self):
        self.assertEqual(
            'systemd-ironic-console-1234',
            self.provider._container_name('1234')
        )

    @mock.patch.object(utils, 'execute', autospec=True)
    def test__reload(self, mock_exec):

        mock_exec.return_value = (None, None)
        self.provider._reload()

        # assert successful call
        mock_exec.assert_called_once_with(
            'systemctl', '--user', 'daemon-reload')

        mock_exec.side_effect = [
            processutils.ProcessExecutionError(
                stderr='ouch'
            ),
            (None, None)
        ]
        # assert failed call
        self.assertRaisesRegex(exception.ConsoleContainerError, 'ouch',
                               self.provider._reload)

    @mock.patch.object(utils, 'execute', autospec=True)
    def test__start(self, mock_exec):

        mock_exec.return_value = (None, None)
        unit = self.provider._unit_name('1234')
        self.provider._start(unit)

        # assert successful call
        mock_exec.assert_called_once_with(
            'systemctl', '--user', 'start', unit)

        mock_exec.side_effect = [
            processutils.ProcessExecutionError(
                stderr='ouch'
            ),
            (None, None)
        ]
        # assert failed call
        self.assertRaisesRegex(exception.ConsoleContainerError, 'ouch',
                               self.provider._start, unit)

    @mock.patch.object(utils, 'execute', autospec=True)
    def test__stop(self, mock_exec):

        mock_exec.return_value = (None, None)
        unit = self.provider._unit_name('1234')
        self.provider._stop(unit)

        # assert successful call
        mock_exec.assert_called_once_with('systemctl', '--user', 'stop', unit)

        mock_exec.side_effect = [
            processutils.ProcessExecutionError(
                stderr='ouch'
            ),
            (None, None)
        ]
        # assert failed call
        self.assertRaisesRegex(exception.ConsoleContainerError, 'ouch',
                               self.provider._stop, unit)

    @mock.patch.object(utils, 'execute', autospec=True)
    def test__host_port(self, mock_exec):

        mock_exec.return_value = ('5900/tcp -> 192.0.2.1:33819', None)
        container = self.provider._container_name('1234')
        self.assertEqual(
            ('192.0.2.1', 33819),
            self.provider._host_port(container)
        )

        # assert successful call
        mock_exec.assert_called_once_with('podman', 'port', container)

        # assert failed parsing response
        mock_exec.return_value = ('5900/tcp -> asdkljffo872', None)
        self.assertRaisesRegex(exception.ConsoleContainerError,
                               'Could not detect port',
                               self.provider._host_port, container)

        mock_exec.side_effect = [
            processutils.ProcessExecutionError(
                stderr=f'Error: no container with name or ID "{container}" '
                       'found: no such container'
            ),
            (None, None)
        ]
        # assert failed call
        self.assertRaisesRegex(exception.ConsoleContainerError,
                               'no such container',
                               self.provider._host_port, container)

    def test__write_container_file(self):
        CONF.set_override(
            'systemd_container_publish_port',
            '192.0.2.2::5900',
            group='vnc')
        CONF.set_override(
            'console_image',
            'localhost/ironic-vnc-container',
            group='vnc')
        CONF.set_override(
            'read_only',
            True,
            group='vnc')

        uuid = '1234'
        container_path = self.provider._container_path(uuid)
        self.provider._write_container_file(
            identifier=uuid, app_name='fake', app_info={})

        # assert the file is correct
        with open(container_path, "r") as f:
            self.assertEqual(
                """[Unit]
Description=A VNC server which displays a console for node 1234

[Container]
Image=localhost/ironic-vnc-container
PublishPort=192.0.2.2::5900
Environment=APP=fake
Environment=APP_INFO='{}'
Environment=READ_ONLY=True

[Install]
WantedBy=default.target""", f.read())

    def test_delete_container_file(self):
        uuid = '1234'
        self.provider._write_container_file(
            uuid, app_name='fake', app_info={})

        container_path = self.provider._container_path(uuid)

        # initial state file exists
        self.assertTrue(os.path.isfile(container_path))

        self.provider._delete_container_file(uuid)

        # assert file was deleted
        self.assertFalse(os.path.exists(container_path))

    @mock.patch.object(utils, 'execute', autospec=True)
    def test_start_stop_container(self, mock_exec):
        uuid = '1234'
        task = mock.Mock(node=mock.Mock(uuid=uuid))

        container_path = self.provider._container_path(uuid)

        mock_exec.side_effect = [
            (None, None),
            (None, None),
            ('5900/tcp -> 192.0.2.1:33819', None)
        ]

        # start the container and assert the host / port
        self.assertEqual(
            ('192.0.2.1', 33819),
            self.provider.start_container(task, 'fake', {})
        )
        # assert the created file
        self.assertTrue(os.path.isfile(container_path))

        # assert all the expected calls
        mock_exec.assert_has_calls([
            mock.call('systemctl', '--user', 'daemon-reload'),
            mock.call('systemctl', '--user', 'start',
                      'ironic-console-1234.service'),
            mock.call('podman', 'port', 'systemd-ironic-console-1234')
        ])

        mock_exec.reset_mock()
        mock_exec.side_effect = [
            (None, None),
            (None, None),
        ]
        # stop the container
        self.provider.stop_container(task)

        # assert the container file is deleted
        self.assertFalse(os.path.exists(container_path))

        # assert expected stop calls
        mock_exec.assert_has_calls([
            mock.call('systemctl', '--user', 'stop',
                      'ironic-console-1234.service'),
            mock.call('systemctl', '--user', 'daemon-reload'),
        ])

    @mock.patch.object(utils, 'execute', autospec=True)
    def test_stop_all_containers(self, mock_exec):
        # set up initial state with 3 running containers
        t1 = mock.Mock(node=mock.Mock(uuid='1234'))
        t2 = mock.Mock(node=mock.Mock(uuid='asdf'))
        t3 = mock.Mock(node=mock.Mock(uuid='foobar'))
        mock_exec.side_effect = [
            (None, None),
            (None, None),
            ('5900/tcp -> 192.0.2.1:33819', None),
            (None, None),
            (None, None),
            ('5900/tcp -> 192.0.2.1:33820', None),
            (None, None),
            (None, None),
            ('5900/tcp -> 192.0.2.1:33821', None),
        ]
        self.provider.start_container(t1, 'fake', {})
        self.provider.start_container(t2, 'fake', {})
        self.provider.start_container(t3, 'fake', {})

        mock_exec.reset_mock()
        mock_exec.side_effect = [
            (None, None),
            (None, None),
            (None, None),
            (None, None),
        ]

        self.provider.stop_all_containers()
        # assert all containers stopped
        mock_exec.assert_has_calls([
            mock.call('systemctl', '--user', 'stop',
                      'ironic-console-1234.service'),
        ])
        mock_exec.assert_has_calls([
            mock.call('systemctl', '--user', 'stop',
                      'ironic-console-asdf.service'),
        ])
        mock_exec.assert_has_calls([
            mock.call('systemctl', '--user', 'stop',
                      'ironic-console-foobar.service'),
        ])
        mock_exec.assert_has_calls([
            mock.call('systemctl', '--user', 'daemon-reload')
        ])

        # stop all containers again and confirm nothing was stopped because
        # all of the files are deleted
        mock_exec.reset_mock()
        self.provider.stop_all_containers()
        mock_exec.assert_not_called()



#!/usr/bin/env python3

import json
import os
import requests
from requests import auth
import signal
import sys
import time
from urllib import parse as urlparse

from selenium import webdriver
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.common.by import By
from selenium.common import exceptions


class BaseApp:

    def __init__(self, app_info):
        self.app_info = app_info

    @property
    def url(self):
        pass

    def handle_exit(self, signum, frame):
        print("got SIGTERM, quitting")
        self.driver.quit()
        sys.exit(0)

    def start(self, driver):
        self.driver = driver
        signal.signal(signal.SIGTERM, self.handle_exit)


class FakeApp(BaseApp):

    @property
    def url(self):
        return "file:///drivers/fake/index.html"


class RedfishApp(BaseApp):

    @property
    def base_url(self):
        return self.app_info["address"]

    @property
    def redfish_url(self):
        return self.base_url + self.app_info.get("root_prefix", "/redfish/v1")

    def disable_right_click(self, driver):
        # disable right-click menu
        driver.execute_script(
            'window.addEventListener("contextmenu", function(e) '
            "{ e.preventDefault(); })"
        )


class IdracApp(RedfishApp):

    @property
    def url(self):
        username = self.app_info["username"]
        password = self.app_info["password"]
        verify = self.app_info.get("verify_ca", True)
        kvm_session_url = (f"{self.redfish_url}/Managers/iDRAC.Embedded.1/Oem/"
                           "Dell/DelliDRACCardService/Actions/DelliDRACCardService.GetKVMSession")
        netloc = urlparse.urlparse(self.base_url).netloc

        r = requests.post(
            kvm_session_url,
            verify=verify,
            timeout=60,
            auth=auth.HTTPBasicAuth(username, password),
            json={"SessionTypeName": "idrac-graphical"},
        ).json()
        temp_username = r["TempUsername"]
        temp_password = r["TempPassword"]
        url = (f"{self.base_url}/restgui/vconsole/index.html?ip={netloc}&"
               f"kvmport=443&title=idrac-graphical&VCSID={temp_username}&VCSID2={temp_password}")
        return url

    def start(self, driver):
        super(IdracApp, self).start(driver)
        # wait for the full screen button
        wait = WebDriverWait(
            driver,
            timeout=10,
            poll_frequency=0.2,
            ignored_exceptions=[exceptions.NoSuchElementException],
        )
        wait.until(
            lambda d: driver.find_element(By.TAG_NAME, value="full-screen")
            or True
        )
        fs_tag = driver.find_element(By.TAG_NAME, value="full-screen")
        fs_tag.find_element(By.TAG_NAME, "button").click()


class IloApp(RedfishApp):

    @property
    def url(self):
        return self.base_url + "/irc.html"

    def login(self, driver):

        username = self.app_info["username"]
        password = self.app_info["password"]
        # wait for the username field to be enabled then perform login
        wait = WebDriverWait(
            driver,
            timeout=10,
            poll_frequency=0.2,
            ignored_exceptions=[exceptions.NoSuchElementException],
        )
        wait.until(
            lambda d: driver.find_element(By.ID, value="username") or True
        )

        username_field = driver.find_element(By.ID, value="username")
        wait = WebDriverWait(
            driver,
            timeout=5,
            poll_frequency=0.2,
            ignored_exceptions=[exceptions.ElementNotInteractableException],
        )
        wait.until(lambda d: username_field.send_keys(username) or True)

        driver.find_element(By.ID, value="password").send_keys(password)
        driver.find_element(By.ID, value="login-form__submit").click()

    def start(self, driver):
        super(IloApp, self).start(driver)

        # Detect iLO 6 vs 5 based on whether a message box or a login form
        # is presented
        try:
            driver.find_element(By.CLASS_NAME, value="loginBoxRestrictWidth")
            is_ilo6 = True
        except exceptions.NoSuchElementException:
            is_ilo6 = False

        if is_ilo6:
            # iLO 6 has an inline login which matches the main login
            self.login(driver)
            self.disable_right_click(driver)
            self.full_screen(driver)
            return

        # load the main login page
        driver.get(self.base_url)

        # full screen content is shown in an embedded iframe
        iframe = driver.find_element(By.ID, "appFrame")
        driver.switch_to.frame(iframe)

        self.login(driver)

        # wait for <body id="app-container"> to exist, which indicates
        # the login form has submitted and session cookies are now set
        wait = WebDriverWait(
            driver,
            timeout=10,
            poll_frequency=0.2,
            ignored_exceptions=[exceptions.NoSuchElementException],
        )
        wait.until(
            lambda d: driver.find_element(By.ID, value="app-container")
            or True
        )

        # load the actual console
        driver.get(self.url)
        self.disable_right_click(driver)
        self.full_screen(driver)

    def full_screen(self, driver):
        # make console full screen to hide menu
        fs_button = driver.find_element(
            By.CLASS_NAME, value="btnVideoFullScreen"
        )
        wait = WebDriverWait(
            driver,
            timeout=20,
            poll_frequency=0.2,
            ignored_exceptions=[
                exceptions.ElementNotInteractableException,
                exceptions.ElementClickInterceptedException,
            ],
        )
        wait.until(lambda d: fs_button.click() or True)


class SupermicroApp(RedfishApp):

    @property
    def url(self):
        return self.base_url

    def start(self, driver):
        super(SupermicroApp, self).start(driver)
        username = self.app_info["username"]
        password = self.app_info["password"]

        # populate login and submit
        driver.find_element(By.NAME, value="name").send_keys(username)
        driver.find_element(By.ID, value="pwd").send_keys(password)
        driver.find_element(By.ID, value="login_word").click()

        # navigate down some iframes
        iframe = driver.find_element(By.ID, "TOPMENU")
        driver.switch_to.frame(iframe)

        iframe = driver.find_element(By.ID, "frame_main")
        driver.switch_to.frame(iframe)

        wait = WebDriverWait(
            driver,
            timeout=30,
            poll_frequency=0.2,
            ignored_exceptions=[
                exceptions.NoSuchElementException,
                exceptions.ElementNotInteractableException,
            ],
        )
        wait.until(lambda d: driver.find_element(By.ID, value="img1") or True)

        # launch the console by waiting for the console preview image to be
        # loaded and clickable
        def snapshot_wait(d):
            try:
                img1 = driver.find_element(By.ID, value="img1")
            except exceptions.NoSuchElementException:
                print("img1 doesn't exist yet")
                return False

            if "Snapshot" not in img1.get_attribute("src"):
                print("img1 src not a console snapshot yet")
                return False
            if not img1.get_attribute("complete") == "true":
                print("img1 console snapshot not loaded yet")
                return False
            try:
                img1.click()
            except exceptions.ElementNotInteractableException:
                print("img1 not clickable yet")
                return False
            return True

        wait = WebDriverWait(driver, timeout=30, poll_frequency=1)
        wait.until(snapshot_wait)

        # self.disable_right_click(driver)


def start_driver(url, app_info):
    print(f"starting app with url {url}")
    opts = webdriver.ChromeOptions()
    opts.binary_location = "/usr/bin/chromium-browser"
    # opts.enable_bidi = True
    if url:
        opts.add_argument(f"--app={url}")

    verify = app_info.get("verify_ca", True)
    if not verify:
        opts.add_argument("--ignore-certificate-errors")
        opts.add_argument("--ignore-ssl-errors")

    opts.add_argument("--disable-extensions")
    opts.add_argument("--disable-gpu")
    opts.add_argument("--disable-plugins-discovery")

    opts.add_argument("--disable-context-menu")
    opts.add_argument("--no-sandbox")
    opts.add_argument("--disable-dev-shm-usage")

    opts.add_argument("--window-position=0,0")
    opts.add_experimental_option("excludeSwitches", ["enable-automation"])
    if "DISPLAY_WIDTH" in os.environ and "DISPLAY_HEIGHT" in os.environ:
        width = int(os.environ["DISPLAY_WIDTH"])
        height = int(os.environ["DISPLAY_HEIGHT"])
        opts.add_argument(f"--window-size={width},{height}")
    if "CHROME_ARGS" in os.environ:
        for arg in os.environ["CHROME_ARGS"].split(" "):
            opts.add_argument(arg)

    driver = webdriver.Chrome(options=opts)
    driver.delete_all_cookies()
    driver.set_window_position(0, 0)

    return driver


def discover_app(app_name, app_info):
    if app_name == "fake":
        return FakeApp
    if app_name == "redfish-graphical":
        # Make an unauthenticated redfish request
        # to discover which console class to use
        url = app_info["address"] + app_info.get("root_prefix", "/redfish/v1")
        verify = app_info.get("verify_ca", True)
        r = requests.get(url, verify=verify, timeout=60).json()
        oem = ",".join(r["Oem"].keys())
        if "Hpe" in oem:
            return IloApp
        if "Dell" in oem:
            return IdracApp
        if "Supermicro" in oem:
            return SupermicroApp
        raise Exception(f"Unsupported {app_name} vendor {oem}")

    raise Exception(f"Unknown app name {app_name}")


def main():
    app_name = os.environ.get("APP")
    print("got app info " + os.environ.get("APP_INFO"))
    app_info = json.loads(os.environ.get("APP_INFO"))
    app_class = discover_app(app_name, app_info)

    app = app_class(app_info)

    driver = start_driver(url=app.url, app_info=app_info)
    print(f"got driver {driver}")

    print(f"Running app {app_name}")
    app.start(driver)
    while True:
        time.sleep(10)


if __name__ == "__main__":
    sys.exit(main())



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from oslo_log import log as logging

from ironic.common import exception
from ironic.drivers.modules.inspector.hooks import base
from ironic.objects import node_inventory

LOG = logging.getLogger(__name__)


class RaidDeviceHook(base.InspectionHook):
    """Hook for learning the root device after RAID creation.

    This hook can figure out the root device in 2 runs. In the first run, the
    node's inventory is saved as usual, and the hook does not do anything.
    The second run will check the difference between the recently discovered
    block devices (as reported by the inspection results) and the previously
    saved ones (from the previously saved inventory). If there is exactly one
    new block device, its serial number is saved in node.properties under the
    'root_device' key.

    This way, it helps to figure out the root device hint in cases when Ironic
    doesn't have enough information to do so otherwise. One such usecase is
    DRAC RAID configuration, where the BMC doesn't provide any useful
    information about the created RAID disks. Using this hook immediately
    before and after creating the root RAID device will solve the issue of
    root device hints.
    """

    def _get_serials(self, inventory):
        if inventory.get('disks'):
            return [x['serial'] for x in inventory.get('disks')
                    if x.get('serial')]

    def __call__(self, task, inventory, plugin_data):
        node = task.node

        if 'root_device' in node.properties:
            LOG.info('Root device is already known for node %s', node.uuid)
            return

        current_devices = self._get_serials(inventory)
        if not current_devices:
            LOG.warning('No block device information was received from the '
                        'ramdisk for node %s', node.uuid)
            return

        try:
            previous_inventory = node_inventory.NodeInventory.get_by_node_id(
                task.context, node.id)
        except exception.NodeInventoryNotFound:
            LOG.debug('Inventory for node %s not found in the database. Raid '
                      'device hook exiting.', task.node.uuid)
            return
        previous_devices = self._get_serials(previous_inventory.get(
            'inventory_data'))

        # Compare previously discovered devices with the current ones
        new_devices = [device for device in current_devices
                       if device not in previous_devices]
        if len(new_devices) > 1:
            LOG.warning('Root device cannot be identified because multiple '
                        'new devices were found for node %s', node.uuid)
            return
        elif len(new_devices) == 0:
            LOG.warning('No new devices were found for node %s', node.uuid)
            return

        node.set_property('root_device', {'serial': new_devices[0]})
        node.save()
        LOG.info('"root_device" property set for node %s', node.uuid)



# Copyright 2016 Intel Corporation
# Copyright 2013 Hewlett-Packard Development Company, L.P.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg
from oslo_config import types as cfg_types

from ironic.common import boot_modes
from ironic.common.i18n import _


class Octal(cfg_types.Integer):

    def __call__(self, value):
        if isinstance(value, int):
            return value
        else:
            return int(str(value), 8)


opts = [
    cfg.HostAddressOpt('host_ip',
                       default='0.0.0.0',
                       help=_('The IP address or hostname on which ironic-api '
                              'listens.')),
    cfg.PortOpt('port',
                default=6385,
                help=_('The TCP port on which ironic-api listens.')),
    cfg.StrOpt('unix_socket',
               help=_('Unix socket to listen on. Disables host_ip and port.')),
    cfg.Opt('unix_socket_mode', type=Octal(),
            help=_('File mode (an octal number) of the unix socket to '
                   'listen on. Ignored if unix_socket is not set.')),
    cfg.IntOpt('max_limit',
               default=1000,
               mutable=True,
               help=_('The maximum number of items returned in a single '
                      'response from a collection resource.')),
    cfg.StrOpt('public_endpoint',
               mutable=True,
               help=_("Public URL to use when building the links to the API "
                      "resources (for example, \"https://ironic.rocks:6384\")."
                      " If None the links will be built using the request's "
                      "host URL. If the API is operating behind a proxy, you "
                      "will want to change this to represent the proxy's URL. "
                      "Defaults to None. "
                      "Ignored when proxy headers parsing is enabled via "
                      "[oslo_middleware]enable_proxy_headers_parsing option.")
               ),
    cfg.IntOpt('api_workers',
               help=_('Number of workers for OpenStack Ironic API service. '
                      'The default is equal to the number of CPUs available, '
                      'but not more than 4. One worker is used if the CPU '
                      'number cannot be detected.')),
    cfg.BoolOpt('enable_ssl_api',
                default=False,
                help=_("Enable the integrated stand-alone API to service "
                       "requests via HTTPS instead of HTTP. If there is a "
                       "front-end service performing HTTPS offloading from "
                       "the service, this option should be False; note, you "
                       "will want to enable proxy headers parsing with "
                       "[oslo_middleware]enable_proxy_headers_parsing "
                       "option or configure [api]public_endpoint option "
                       "to set URLs in responses to the SSL terminated one.")),
    cfg.BoolOpt('restrict_lookup',
                default=True,
                mutable=True,
                help=_('Whether to restrict the lookup API to only nodes '
                       'in certain states. Setting this to False can be '
                       'insecure and is not advisable.')),
    cfg.IntOpt('ramdisk_heartbeat_timeout',
               default=300,
               mutable=True,
               help=_('Maximum interval (in seconds) for agent heartbeats.')),
    cfg.StrOpt(
        'network_data_schema',
        default='$pybasedir/api/controllers/v1/network-data-schema.json',
        help=_("Schema for network data used by this deployment.")),
    cfg.BoolOpt('project_admin_can_manage_own_nodes',
                default=True,
                mutable=True,
                help=_('If a project scoped administrative user is permitted '
                       'to create/delete baremetal nodes in their project.')),
    cfg.ListOpt('disallowed_enrollment_boot_modes',
                item_type=cfg_types.String(
                    choices=[
                        (boot_modes.UEFI, _('UEFI boot mode')),
                        (boot_modes.LEGACY_BIOS, _('Legacy BIOS boot mode'))],
                ),
                default=[],
                mutable=True,
                help=_("Specifies a list of boot modes that are not allowed "
                       "during enrollment. Eg: ['bios']")),
]

opt_group = cfg.OptGroup(name='api',
                         title='Options for the ironic-api service')


def register_opts(conf):
    conf.register_group(opt_group)
    conf.register_opts(opts, group=opt_group)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_utils import uuidutils

from ironic.conductor import task_manager
from ironic.conf import CONF
from ironic.drivers.modules.inspector.hooks import ports as ports_hook
from ironic import objects
from ironic.tests.unit.db import base as db_base
from ironic.tests.unit.drivers.modules.inspector.hooks. \
    test_validate_interfaces import _PXE_INTERFACE
from ironic.tests.unit.drivers.modules.inspector.hooks. \
    test_validate_interfaces import _VALID
from ironic.tests.unit.objects import utils as obj_utils


class AddPortsTestCase(db_base.DbTestCase):
    def setUp(self):
        super().setUp()
        CONF.set_override('enabled_inspect_interfaces',
                          ['agent', 'no-inspect'])
        self.node = obj_utils.create_test_node(self.context,
                                               inspect_interface='agent')
        self.interfaces = {key: value.copy() for key, value in _VALID.items()
                           if key not in ('lo', 'em4')}
        self.macs = {
            _PXE_INTERFACE,
            '11:11:11:11:11:11',
            '22:22:22:22:22:22',
            '33:33:33:33:33:33',
        }

    def test_add_ports(self):
        with task_manager.acquire(self.context, self.node.id) as task:
            ports_hook.add_ports(task, self.interfaces)
        ports = objects.Port.list_by_node_id(self.context, self.node.id)
        self.assertEqual(
            {port.address: port.pxe_enabled for port in ports},
            {mac: (mac == _PXE_INTERFACE) for mac in self.macs})

    def test_duplicates(self):
        obj_utils.create_test_port(self.context,
                                   node_id=self.node.id,
                                   address=_PXE_INTERFACE,
                                   pxe_enabled=False)
        with task_manager.acquire(self.context, self.node.id) as task:
            ports_hook.add_ports(task, self.interfaces)
        ports = objects.Port.list_by_node_id(self.context, self.node.id)
        self.assertEqual(
            {port.address: port.pxe_enabled for port in ports},
            # Always False because the PXE port already existed
            {mac: False for mac in self.macs})


class UpdatePortsTestCase(db_base.DbTestCase):
    def setUp(self):
        super().setUp()
        CONF.set_override('enabled_inspect_interfaces',
                          ['agent', 'no-inspect'])
        self.node = obj_utils.create_test_node(self.context,
                                               inspect_interface='agent')
        self.macs = {
            _PXE_INTERFACE,
            '11:11:11:11:11:11',
            '22:22:22:22:22:22',
            '33:33:33:33:33:33',
        }
        self.present_macs = {i['mac_address'] for i in _VALID.values()}
        self.extra_mac = '00:11:00:11:00:11'
        self.all_macs = self.present_macs | {self.extra_mac}
        for mac in self.all_macs:
            obj_utils.create_test_port(self.context,
                                       uuid=uuidutils.generate_uuid(),
                                       node_id=self.node.id,
                                       address=mac,
                                       pxe_enabled=(mac == self.extra_mac))

    def test_keep_all(self):
        with task_manager.acquire(self.context, self.node.id) as task:
            ports_hook.update_ports(task, _VALID, self.macs)
        ports = objects.Port.list_by_node_id(self.context, self.node.id)
        self.assertEqual(
            {port.address: port.pxe_enabled for port in ports},
            # Nothing is removed by default, pxe_enabled updated
            {mac: (mac == _PXE_INTERFACE) for mac in self.all_macs})

    def test_keep_pxe_enabled(self):
        CONF.set_override('update_pxe_enabled', False, group='inspector')
        with task_manager.acquire(self.context, self.node.id) as task:
            ports_hook.update_ports(task, _VALID, self.macs)
        ports = objects.Port.list_by_node_id(self.context, self.node.id)
        self.assertEqual(
            {port.address: port.pxe_enabled for port in ports},
            # Nothing is removed by default, pxe_enabled kept intact
            {mac: (mac == self.extra_mac) for mac in self.all_macs})

    def test_keep_added(self):
        CONF.set_override('keep_ports', 'added', group='inspector')
        with task_manager.acquire(self.context, self.node.id) as task:
            ports_hook.update_ports(task, _VALID, self.macs)
        ports = objects.Port.list_by_node_id(self.context, self.node.id)
        self.assertEqual(
            {port.address: port.pxe_enabled for port in ports},
            # Extra ports removed, pxe_enabled updated
            {mac: (mac == _PXE_INTERFACE) for mac in self.macs})

    def test_keep_present(self):
        CONF.set_override('keep_ports', 'present', group='inspector')
        with task_manager.acquire(self.context, self.node.id) as task:
            ports_hook.update_ports(task, _VALID, self.macs)
        ports = objects.Port.list_by_node_id(self.context, self.node.id)
        self.assertEqual(
            {port.address: port.pxe_enabled for port in ports},
            # Extra port removed, pxe_enabled updated
            {mac: (mac == _PXE_INTERFACE) for mac in self.present_macs})



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Inspection implementation for the conductor."""

from oslo_log import log
from oslo_utils import excutils

from ironic.common import exception
from ironic.common.i18n import _
from ironic.common.inspection_rules import engine
from ironic.common import states
from ironic.conductor import task_manager
from ironic.conductor import utils
from ironic.drivers.modules import inspect_utils

LOG = log.getLogger(__name__)


@task_manager.require_exclusive_lock
def inspect_hardware(task):
    """Initiates inspection.

    :param task: a TaskManager instance with an exclusive lock
                 on its node.
    :raises: HardwareInspectionFailure if driver doesn't
             return the state as states.MANAGEABLE, states.INSPECTWAIT.

    """
    node = task.node

    def handle_failure(e, log_func=LOG.error):
        utils.node_history_record(task.node, event=e,
                                  event_type=states.INTROSPECTION,
                                  error=True, user=task.context.user_id)
        task.process_event('fail')
        log_func("Failed to inspect node %(node)s: %(err)s",
                 {'node': node.uuid, 'err': e})

    # Inspection cannot start in fast-track mode, wipe token and URL.
    utils.wipe_token_and_url(task)

    try:
        new_state = task.driver.inspect.inspect_hardware(task)
    except exception.IronicException as e:
        with excutils.save_and_reraise_exception():
            error = str(e)
            handle_failure(error)
    except Exception as e:
        error = (_('Unexpected exception of type %(type)s: %(msg)s') %
                 {'type': type(e).__name__, 'msg': e})
        handle_failure(error, log_func=LOG.exception)
        raise exception.HardwareInspectionFailure(error=error)

    if new_state == states.MANAGEABLE:
        task.process_event('done')
        LOG.info('Successfully inspected node %(node)s',
                 {'node': node.uuid})
    elif new_state == states.INSPECTWAIT:
        task.process_event('wait')
        LOG.info('Successfully started inspection on node %(node)s',
                 {'node': node.uuid})
    else:
        error = (_("During inspection, driver returned unexpected "
                   "state %(state)s") % {'state': new_state})
        handle_failure(error)
        raise exception.HardwareInspectionFailure(error=error)


@task_manager.require_exclusive_lock
def abort_inspection(task):
    """Abort inspection for the node."""
    node = task.node

    try:
        task.driver.inspect.abort(task)
    except exception.UnsupportedDriverExtension:
        with excutils.save_and_reraise_exception():
            LOG.error('Inspect interface "%(intf)s" does not support abort '
                      'operation for node %(node)s',
                      {'intf': node.inspect_interface, 'node': node.uuid})
    except Exception as e:
        with excutils.save_and_reraise_exception():
            LOG.exception('Error when aborting inspection of node %(node)s',
                          {'node': node.uuid})
            error = _('Failed to abort inspection: %s') % e
            utils.node_history_record(task.node, event=error,
                                      event_type=states.INTROSPECTION,
                                      error=True,
                                      user=task.context.user_id)
            node.save()

    error = _('Inspection was aborted by request.')
    utils.node_history_record(task.node, event=error,
                              event_type=states.INTROSPECTION,
                              error=True,
                              user=task.context.user_id)
    utils.wipe_token_and_url(task)
    if task.node.provision_state != states.INSPECTFAIL:
        task.process_event('abort')
    else:
        task.node.save()
    LOG.info('Successfully aborted inspection of node %(node)s',
             {'node': node.uuid})


@task_manager.require_exclusive_lock
def continue_inspection(task, inventory, plugin_data):
    """Continue inspection for the node."""
    node = task.node
    LOG.debug('Inventory for node %(node)s: %(data)s',
              {'node': node.uuid, 'data': inventory})
    plugin_data = plugin_data or {}

    try:
        result = task.driver.inspect.continue_inspection(
            task, inventory, plugin_data)

        if result == states.INSPECTWAIT:
            if task.node.provision_state == states.INSPECTING:
                task.process_event('wait')
            LOG.debug('Waiting for inspection data to be processed '
                      'asynchronously for node %s', node.uuid)
            return

        engine.apply_rules(task, inventory, plugin_data, 'main')

        # NOTE(dtantsur): logs can be huge and are stored separately
        plugin_data.pop('logs', None)

        # NOTE(dtantsur): this is done *after* processing to allow
        # modifications, especially to plugin_data.
        inspect_utils.store_inspection_data(
            node, inventory, plugin_data, context=task.context)
    except exception.UnsupportedDriverExtension:
        with excutils.save_and_reraise_exception():
            intf_name = task.driver.inspect.__class__.__name__
            LOG.error('Inspect interface %(intf)s does not '
                      'support processing inspection data for node %(node)s',
                      {'intf': intf_name, 'node': node.uuid})
    except Exception as e:
        with excutils.save_and_reraise_exception():
            LOG.exception('Error when processing inspection data for '
                          'node %(node)s', {'node': node.uuid})
            error = _('Failed to finish inspection: %s') % e
            utils.node_history_record(task.node, event=error,
                                      event_type=states.INTROSPECTION,
                                      error=True)
            if node.provision_state != states.ENROLL:
                task.process_event('fail')

    if node.provision_state != states.ENROLL:
        task.process_event('done')
        LOG.info('Successfully finished inspection of node %s', node.uuid)
    else:
        LOG.info('Successfully finished auto-discovery of node %s', node.uuid)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""
Tests for the API /inspection_rules methods.
"""
import datetime
from http import client as http_client
from unittest import mock

from oslo_utils import timeutils
from oslo_utils import uuidutils

from ironic.api.controllers import base as api_base
from ironic.api.controllers import v1 as api_v1
from ironic.api.controllers.v1 import notification_utils
from ironic import objects
from ironic.objects import fields as obj_fields
from ironic.tests.unit.api import base as test_api_base
from ironic.tests.unit.api import utils as test_api_utils
from ironic.tests.unit.objects import utils as obj_utils


class BaseInspectionRulesAPITest(test_api_base.BaseApiTest):
    headers = {api_base.Version.string: str(api_v1.max_version())}
    invalid_version_headers = {api_base.Version.string: '1.92'}


class TestListInspectionRules(BaseInspectionRulesAPITest):

    def test_empty(self):
        data = self.get_json('/inspection_rules', headers=self.headers)
        self.assertEqual([], data['inspection_rules'])

    def test_one(self):
        inspection_rule = obj_utils.create_test_inspection_rule(self.context)
        data = self.get_json('/inspection_rules', headers=self.headers)
        self.assertEqual(1, len(data['inspection_rules']))
        self.assertEqual(inspection_rule.uuid,
                         data['inspection_rules'][0]['uuid'])
        self.assertEqual(inspection_rule.description,
                         data['inspection_rules'][0]['description'])
        self.assertNotIn('actions', data['inspection_rules'][0])
        self.assertNotIn('conditions', data['inspection_rules'][0])

    def test_get_one(self):
        rule = obj_utils.create_test_inspection_rule(self.context)
        data = self.get_json('/inspection_rules/%s' % rule.uuid,
                             headers=self.headers)
        self.assertEqual(rule.uuid, data['uuid'])
        self.assertIn('conditions', data)
        self.assertIn('actions', data)

    def test_get_rule_data(self):
        """Test get normal rule does not hide conditions and actions"""
        idict = test_api_utils.post_get_test_inspection_rule()
        idict['sensitive'] = False
        idict['conditions'] = [{'op': 'eq', 'args': {'values': [1, 1]}}]
        idict['actions'] = [{'op': 'set-attribute',
                             'args': {'path': 'test', 'value': 'secret'}}]

        response = self.post_json('/inspection_rules', idict,
                                  headers=self.headers)
        self.assertEqual(201, response.status_int)

        rule = self.get_json('/inspection_rules/%s' % idict['uuid'],
                             headers=self.headers)
        self.assertFalse(rule['sensitive'])
        self.assertIsNotNone(rule['conditions'])
        self.assertIsNotNone(rule['actions'])

    def test_get_sensitive_rule_hides_data(self):
        """Test get sensitive rule hides conditions and actions"""
        idict = test_api_utils.post_get_test_inspection_rule()
        idict['sensitive'] = True
        idict['conditions'] = [{'op': 'eq', 'args': {'values': [1, 1]}}]
        idict['actions'] = [{'op': 'set-attribute',
                             'args': {'path': 'test', 'value': 'secret'}}]

        response = self.post_json('/inspection_rules', idict,
                                  headers=self.headers)
        self.assertEqual(201, response.status_int)

        rule = self.get_json('/inspection_rules/%s' % idict['uuid'],
                             headers=self.headers)
        self.assertTrue(rule['sensitive'])
        self.assertIsNone(rule['conditions'])
        self.assertIsNone(rule['actions'])

    def test_list_hides_sensitive_data(self):
        """Test that listing rules hides sensitive data for sensitive rules."""
        sensitive_rule = test_api_utils.post_get_test_inspection_rule()
        sensitive_rule['sensitive'] = True
        sensitive_rule['uuid'] = uuidutils.generate_uuid()

        normal_rule = test_api_utils.post_get_test_inspection_rule()
        normal_rule['sensitive'] = False
        normal_rule['uuid'] = uuidutils.generate_uuid()

        self.post_json('/inspection_rules', sensitive_rule,
                       headers=self.headers)
        self.post_json('/inspection_rules', normal_rule, headers=self.headers)

        data = self.get_json('/inspection_rules?detail=true',
                             headers=self.headers)
        sensitive_result = next(r for r in data['inspection_rules']
                                if r['uuid'] == sensitive_rule['uuid'])
        normal_result = next(r for r in data['inspection_rules']
                             if r['uuid'] == normal_rule['uuid'])

        self.assertTrue(sensitive_result['sensitive'])
        self.assertIsNone(sensitive_result['conditions'])
        self.assertIsNone(sensitive_result['actions'])

        self.assertFalse(normal_result['sensitive'])
        self.assertIsNotNone(normal_result['conditions'])
        self.assertIsNotNone(normal_result['actions'])

    def test_get_all_invalid_api_version(self):
        obj_utils.create_test_inspection_rule(self.context)
        response = self.get_json('/inspection_rules',
                                 headers=self.invalid_version_headers,
                                 expect_errors=True)
        self.assertEqual(http_client.NOT_FOUND, response.status_int)

    def test_get_one_invalid_api_version(self):
        inspection_rule = obj_utils.create_test_inspection_rule(self.context)
        response = self.get_json(
            '/inspection_rules/%s' % (inspection_rule.uuid),
            headers=self.invalid_version_headers,
            expect_errors=True)
        self.assertEqual(http_client.NOT_FOUND, response.status_int)

    def test_get_all(self):
        obj_utils.create_test_inspection_rule(self.context)
        obj_utils.create_test_inspection_rule(self.context)
        data = self.get_json('/inspection_rules', headers=self.headers)
        self.assertEqual(2, len(data['inspection_rules']))


class TestPost(BaseInspectionRulesAPITest):
    @mock.patch.object(timeutils, 'utcnow', autospec=True)
    def test_create_rule(self, mock_utcnow):
        idict = test_api_utils.post_get_test_inspection_rule()
        test_time = datetime.datetime(2024, 8, 27, 0, 0)
        mock_utcnow.return_value = test_time

        response = self.post_json('/inspection_rules', idict,
                                  headers=self.headers)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(201, response.status_int)
        result = self.get_json('/inspection_rules/%s' % idict['uuid'],
                               headers=self.headers)
        self.assertEqual(idict['uuid'], result['uuid'])
        self.assertFalse(result['updated_at'])
        return_created_at = timeutils.parse_isotime(
            result['created_at']).replace(tzinfo=None)
        self.assertEqual(test_time, return_created_at)
        # Check location header
        self.assertIsNotNone(response.location)
        expected_location = '/v1/inspection_rules/%s' % idict['uuid']
        self.assertEqual(expected_location,
                         response.location[response.location.index('/v1'):])

    def test_create_rule_generate_uuid(self):
        idict = test_api_utils.post_get_test_inspection_rule()
        del idict['uuid']
        response = self.post_json('/inspection_rules', idict,
                                  headers=self.headers)
        result = self.get_json('/inspection_rules/%s' % response.json['uuid'],
                               headers=self.headers)
        self.assertTrue(uuidutils.is_uuid_like(result['uuid']))
        self.assertNotIn('id', result)

    def test_create_rule_with_optional_args(self):
        idict = test_api_utils.post_get_test_inspection_rule()
        idict['conditions'] = [
            {'op': 'eq', 'args': {'values': [5, 5]}, 'force_strings': True},
            {'op': 'gt', 'args': {'values': [10, 5]}}
        ]
        idict['actions'] = [
            {'op': 'extend-attribute', 'args': {
                'path': 'properties/capabilities', 'value': 'test:value'},
                'unique': True},
            {'op': 'set-attribute', 'args': {
                'path': 'properties/test', 'value': 'test-value'}}
        ]

        response = self.post_json('/inspection_rules', idict,
                                  headers=self.headers)
        self.assertEqual(201, response.status_int)

    def test_create_rule_with_invalid_priority_fails(self):
        idict = test_api_utils.post_get_test_inspection_rule()
        idict['priority'] = -1
        response = self.post_json('/inspection_rules', idict,
                                  headers=self.headers,
                                  expect_errors=True)
        self.assertEqual(http_client.BAD_REQUEST, response.status_int)


class TestPatch(BaseInspectionRulesAPITest):
    def test_patch_invalid_api_version(self):
        rule = obj_utils.create_test_inspection_rule(self.context)
        patch = [{'op': 'replace', 'path': '/description',
                  'value': 'New description'}]

        response = self.patch_json('/inspection_rules/%s' % rule.uuid,
                                   patch, headers=self.invalid_version_headers,
                                   expect_errors=True)
        self.assertEqual(http_client.METHOD_NOT_ALLOWED, response.status_int)

    def test_set_sensitive_field(self):
        idict = test_api_utils.post_get_test_inspection_rule()
        idict['sensitive'] = False

        response = self.post_json('/inspection_rules', idict,
                                  headers=self.headers)
        self.assertEqual(201, response.status_int)

        # A non-sensitive rule can be marked sensitive, but not if already set
        patch = [{'op': 'replace', 'path': '/sensitive', 'value': True}]
        response = self.patch_json(
            '/inspection_rules/%s' % idict['uuid'],
            patch,
            headers=self.headers,
            expect_errors=True
        )

        # Should succeed
        self.assertEqual(http_client.OK, response.status_int)

        # Should fail
        new_patch = [{'op': 'replace', 'path': '/sensitive', 'value': False}]
        response = self.patch_json(
            '/inspection_rules/%s' % idict['uuid'],
            new_patch,
            headers=self.headers,
            expect_errors=True
        )
        self.assertEqual(http_client.BAD_REQUEST, response.status_int)


@mock.patch.object(objects.InspectionRule, 'destroy', autospec=True)
class TestDelete(BaseInspectionRulesAPITest):

    @mock.patch.object(notification_utils, '_emit_api_notification',
                       autospec=True)
    def test_delete_by_uuid(self, mock_notify, mock_destroy):
        rule = obj_utils.create_test_inspection_rule(self.context)
        self.delete('/inspection_rules/%s' % rule.uuid,
                    headers=self.headers)
        mock_destroy.assert_called_once_with(mock.ANY)
        mock_notify.assert_has_calls([mock.call(mock.ANY, mock.ANY, 'delete',
                                      obj_fields.NotificationLevel.INFO,
                                      obj_fields.NotificationStatus.START),
                                      mock.call(mock.ANY, mock.ANY, 'delete',
                                      obj_fields.NotificationLevel.INFO,
                                      obj_fields.NotificationStatus.END)])

    def test_delete_invalid_api_version(self, mock_destroy):
        rule = obj_utils.create_test_inspection_rule(self.context)
        response = self.delete(
            '/inspection_rules/%s' % rule.uuid,
            expect_errors=True,
            headers=self.invalid_version_headers)
        self.assertEqual(http_client.METHOD_NOT_ALLOWED, response.status_int)



# coding=utf-8

# Copyright 2024 Red Hat, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from oslo_config import cfg
from oslo_utils import fileutils

from ironic.common import checksum_utils
from ironic.common import exception
from ironic.common import image_service
from ironic.tests import base

CONF = cfg.CONF


@mock.patch.object(checksum_utils, 'compute_image_checksum',
                   autospec=True)
class IronicChecksumUtilsValidateTestCase(base.TestCase):

    def test_validate_checksum(self, mock_compute):
        mock_compute.return_value = 'f00'
        checksum_utils.validate_checksum('path', 'f00', 'algo')
        mock_compute.assert_called_once_with('path', 'algo')

    def test_validate_checksum_mixed_case(self, mock_compute):
        mock_compute.return_value = 'f00'
        checksum_utils.validate_checksum('path', 'F00', 'ALGO')
        mock_compute.assert_called_once_with('path', 'algo')

    def test_validate_checksum_mixed_md5(self, mock_compute):
        mock_compute.return_value = 'f00'
        checksum_utils.validate_checksum('path', 'F00')
        mock_compute.assert_called_once_with('path')

    def test_validate_checksum_mismatch(self, mock_compute):
        mock_compute.return_value = 'a00'
        self.assertRaises(exception.ImageChecksumError,
                          checksum_utils.validate_checksum,
                          'path', 'f00', 'algo')
        mock_compute.assert_called_once_with('path', 'algo')

    def test_validate_checksum_hashlib_not_supports_algo(self, mock_compute):
        mock_compute.side_effect = ValueError()
        self.assertRaises(exception.ImageChecksumAlgorithmFailure,
                          checksum_utils.validate_checksum,
                          'path', 'f00', 'algo')
        mock_compute.assert_called_once_with('path', 'algo')

    def test_validate_checksum_file_not_found(self, mock_compute):
        mock_compute.side_effect = OSError()
        self.assertRaises(exception.ImageChecksumFileReadFailure,
                          checksum_utils.validate_checksum,
                          'path', 'f00', 'algo')
        mock_compute.assert_called_once_with('path', 'algo')

    def test_validate_checksum_mixed_case_delimited(self, mock_compute):
        mock_compute.return_value = 'f00'
        checksum_utils.validate_checksum('path', 'algo:F00')
        mock_compute.assert_called_once_with('path', 'algo')


class IronicChecksumUtilsTestCase(base.TestCase):

    def test_is_checksum_url_string(self):
        self.assertFalse(checksum_utils.is_checksum_url('f00'))

    def test_is_checksum_url_file(self):
        self.assertFalse(checksum_utils.is_checksum_url('file://foo'))

    def test_is_checksum_url(self):
        urls = ['http://foo.local/file',
                'https://foo.local/file']
        for url in urls:
            self.assertTrue(checksum_utils.is_checksum_url(url))

    def test_get_checksum_and_algo_image_checksum(self):
        value = 'c46f2c98efe1cd246be1796cd842246e'
        i_info = {'image_checksum': value}
        csum, algo = checksum_utils.get_checksum_and_algo(i_info)
        self.assertEqual(value, csum)
        self.assertIsNone(algo)

    def test_get_checksum_and_algo_image_checksum_not_allowed(self):
        CONF.set_override('allow_md5_checksum', False, group='agent')
        value = 'c46f2c98efe1cd246be1796cd842246e'
        i_info = {'image_checksum': value}
        self.assertRaises(exception.ImageChecksumAlgorithmFailure,
                          checksum_utils.get_checksum_and_algo,
                          i_info)

    def test_get_checksum_and_algo_image_checksum_glance(self):
        value = 'c46f2c98efe1cd246be1796cd842246e'
        i_info = {'image_os_hash_value': value,
                  'image_os_hash_algo': 'foobar'}
        csum, algo = checksum_utils.get_checksum_and_algo(i_info)
        self.assertEqual(value, csum)
        self.assertEqual('foobar', algo)

    def test_get_checksum_and_algo_image_checksum_sha256(self):
        value = 'a' * 64
        i_info = {'image_checksum': value}
        csum, algo = checksum_utils.get_checksum_and_algo(i_info)
        self.assertEqual(value, csum)
        self.assertEqual('sha256', algo)

    def test_get_checksum_and_algo_image_checksum_sha512(self):
        value = 'f' * 128
        i_info = {'image_checksum': value}
        csum, algo = checksum_utils.get_checksum_and_algo(i_info)
        self.assertEqual(value, csum)
        self.assertEqual('sha512', algo)

    @mock.patch.object(checksum_utils, 'get_checksum_from_url', autospec=True)
    def test_get_checksum_and_algo_image_checksum_http_url(self, mock_get):
        value = 'http://checksum-url'
        i_info = {
            'image_checksum': value,
            'image_source': 'image-ref'
        }
        mock_get.return_value = 'f' * 64
        csum, algo = checksum_utils.get_checksum_and_algo(i_info)
        mock_get.assert_called_once_with(value, 'image-ref')
        self.assertEqual('f' * 64, csum)
        self.assertEqual('sha256', algo)

    @mock.patch.object(checksum_utils, 'get_checksum_from_url', autospec=True)
    def test_get_checksum_and_algo_image_checksum_https_url(self, mock_get):
        value = 'https://checksum-url'
        i_info = {
            'image_checksum': value,
            'image_source': 'image-ref'
        }
        mock_get.return_value = 'f' * 128
        csum, algo = checksum_utils.get_checksum_and_algo(i_info)
        mock_get.assert_called_once_with(value, 'image-ref')
        self.assertEqual('f' * 128, csum)
        self.assertEqual('sha512', algo)

    @mock.patch.object(fileutils, 'compute_file_checksum', autospec=True)
    def test_get_checksum_and_algo_no_checksum_file_url(self, mock_cfc):
        i_info = {
            'image_source': 'file:///var/lib/ironic/images/foo.raw'
        }
        mock_cfc.return_value = 'f' * 64
        csum, algo = checksum_utils.get_checksum_and_algo(i_info)
        mock_cfc.assert_called_once_with('/var/lib/ironic/images/foo.raw',
                                         algorithm='sha256')
        self.assertEqual('f' * 64, csum)
        self.assertEqual('sha256', algo)

    def test_validate_text_checksum(self):
        csum = ('sha256:02edbb53017ded13c286e27d14285cb82f5a'
                '87f6dcbae280d6c53b5d98477bb7')
        res = checksum_utils.validate_text_checksum('me0w', csum)
        self.assertIsNone(res)

    def test_validate_text_checksum_invalid(self):
        self.assertRaises(exception.ImageChecksumError,
                          checksum_utils.validate_text_checksum,
                          'me0w', 'sha256:f00')


@mock.patch.object(image_service.HttpImageService, 'get',
                   autospec=True)
class IronicChecksumUtilsGetChecksumTestCase(base.TestCase):

    def test_get_checksum_from_url_empty_response(self, mock_get):
        mock_get.return_value = ''
        error = ('Failed to download image https://checksum-url, '
                 'reason: Checksum file empty.')
        self.assertRaisesRegex(exception.ImageDownloadFailed,
                               error,
                               checksum_utils.get_checksum_from_url,
                               'https://checksum-url',
                               'https://image-url/file')
        mock_get.assert_called_once_with('https://checksum-url')

    def test_get_checksum_from_url_one_line(self, mock_get):
        mock_get.return_value = 'a' * 32
        csum = checksum_utils.get_checksum_from_url(
            'https://checksum-url', 'https://image-url/file')
        mock_get.assert_called_once_with('https://checksum-url')
        self.assertEqual('a' * 32, csum)

    def test_get_checksum_from_url_nomatch_line(self, mock_get):
        mock_get.return_value = 'foobar'
        # For some reason assertRaisesRegex really doesn't like
        # the error. Easiest path is just to assertTrue the compare.
        exc = self.assertRaises(exception.ImageDownloadFailed,
                                checksum_utils.get_checksum_from_url,
                                'https://checksum-url',
                                'https://image-url/file')
        self.assertTrue(
            'Invalid checksum file (No valid checksum found' in str(exc))
        mock_get.assert_called_once_with('https://checksum-url')

    def test_get_checksum_from_url_multiline(self, mock_get):
        test_csum = ('f2ca1bb6c7e907d06dafe4687e579fce76b37e4e9'
                     '3b7605022da52e6ccc26fd2')
        mock_get.return_value = ('fee f00\n%s file\nbar fee\nf00' % test_csum)
        # For some reason assertRaisesRegex really doesn't like
        # the error. Easiest path is just to assertTrue the compare.
        checksum = checksum_utils.get_checksum_from_url(
            'https://checksum-url',
            'https://image-url/file')
        self.assertEqual(test_csum, checksum)
        mock_get.assert_called_once_with('https://checksum-url')

    def test_get_checksum_from_url_multiline_no_file(self, mock_get):
        test_csum = 'a' * 64
        error = ("Failed to download image https://checksum-url, reason: "
                 "Checksum file does not contain name file")
        mock_get.return_value = ('f00\n%s\nbar\nf00' % test_csum)
        # For some reason assertRaisesRegex really doesn't like
        # the error. Easiest path is just to assertTrue the compare.
        self.assertRaisesRegex(exception.ImageDownloadFailed,
                               error,
                               checksum_utils.get_checksum_from_url,
                               'https://checksum-url',
                               'https://image-url/file')
        mock_get.assert_called_once_with('https://checksum-url')



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_versionedobjects import base as object_base

from ironic.db import api as dbapi
from ironic.objects import base
from ironic.objects import fields as object_fields


@base.IronicObjectRegistry.register
class NodeInventory(base.IronicObject, object_base.VersionedObjectDictCompat):
    # Version 1.0: Initial version
    VERSION = '1.0'

    dbapi = dbapi.get_instance()

    fields = {
        'id': object_fields.IntegerField(),
        'node_id': object_fields.IntegerField(nullable=True),
        'inventory_data': object_fields.FlexibleDictField(nullable=True),
        'plugin_data': object_fields.FlexibleDictField(nullable=True),
    }

    @classmethod
    def _from_node_object(cls, context, node):
        """Convert a node into a virtual `NodeInventory` object."""
        result = cls(context)
        result._update_from_node_object(node)
        return result

    def _update_from_node_object(self, node):
        """Update the NodeInventory object from the node."""
        for src, dest in self.node_mapping.items():
            setattr(self, dest, getattr(node, src, None))
        for src, dest in self.instance_info_mapping.items():
            setattr(self, dest, node.instance_info.get(src))

    @classmethod
    def get_by_node_id(cls, context, node_id):
        """Get a NodeInventory object by its node ID.

        :param cls: the :class:`NodeInventory`
        :param context: Security context
        :param uuid: The UUID of a NodeInventory.
        :returns: A :class:`NodeInventory` object.
        :raises: NodeInventoryNotFound

        """
        db_inventory = cls.dbapi.get_node_inventory_by_node_id(node_id)
        inventory = cls._from_db_object(context, cls(), db_inventory)
        return inventory

    def create(self, context=None):
        """Create a NodeInventory record in the DB.

        :param context: Security context. NOTE: This should only
                        be used internally by the indirection_api.
                        Unfortunately, RPC requires context as the first
                        argument, even though we don't use it.
                        A context should be set when instantiating the
                        object, e.g.: NodeHistory(context)
        """
        values = self.do_version_changes_for_db()
        db_inventory = self.dbapi.create_node_inventory(values)
        self._from_db_object(self._context, self, db_inventory)

    def destroy(self, context=None):
        """Delete the NodeInventory from the DB.

        :param context: Security context. NOTE: This should only
                        be used internally by the indirection_api.
                        Unfortunately, RPC requires context as the first
                        argument, even though we don't use it.
                        A context should be set when instantiating the
                        object, e.g.: NodeInventory(context)
        :raises: NodeInventoryNotFound
        """
        self.dbapi.destroy_node_inventory_by_node_id(self.node_id)
        self.obj_reset_changes()



# Copyright 2013 Hewlett-Packard Development Company, L.P.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from http import client as http_client
from unittest import mock

from webob import exc

from ironic.api.controllers import base as cbase
from ironic.tests.unit.api import base


class TestBase(base.BaseApiTest):

    def test_api_setup(self):
        pass

    def test_bad_uri(self):
        response = self.get_json('/bad/path',
                                 expect_errors=True,
                                 headers={"Accept": "application/json"})
        self.assertEqual(http_client.NOT_FOUND, response.status_int)
        self.assertEqual("application/json", response.content_type)
        self.assertTrue(response.json['error_message'])


class TestVersion(base.BaseApiTest):

    @mock.patch('ironic.api.controllers.base.Version.parse_headers',
                autospec=True)
    def test_init(self, mock_parse):
        a = mock.Mock()
        b = mock.Mock()
        mock_parse.return_value = (a, b)
        v = cbase.Version('test', 'foo', 'bar')

        mock_parse.assert_called_with('test', 'foo', 'bar')
        self.assertEqual(a, v.major)
        self.assertEqual(b, v.minor)

    @mock.patch('ironic.api.controllers.base.Version.parse_headers',
                autospec=True)
    def test_repr(self, mock_parse):
        mock_parse.return_value = (123, 456)
        v = cbase.Version('test', mock.ANY, mock.ANY)
        result = "%s" % v
        self.assertEqual('123.456', result)

    @mock.patch('ironic.api.controllers.base.Version.parse_headers',
                autospec=True)
    def test_repr_with_strings(self, mock_parse):
        mock_parse.return_value = ('abc', 'def')
        v = cbase.Version('test', mock.ANY, mock.ANY)
        result = "%s" % v
        self.assertEqual('abc.def', result)

    def test_parse_headers_ok(self):
        version = cbase.Version.parse_headers(
            {cbase.Version.string: '123.456'}, mock.ANY, mock.ANY)
        self.assertEqual((123, 456), version)

    def test_parse_new_standard_singular_header_ok(self):
        version = cbase.Version.parse_headers(
            {'OpenStack-API-Version': 'baremetal 123.456'}, mock.ANY, mock.ANY)
        self.assertEqual((123, 456), version)

    def test_parse_headers_latest(self):
        for s in ['latest', 'LATEST']:
            version = cbase.Version.parse_headers(
                {cbase.Version.string: s}, mock.ANY, '1.9')
            self.assertEqual((1, 9), version)

    def test_parse_headers_bad_length(self):
        self.assertRaises(
            exc.HTTPNotAcceptable,
            cbase.Version.parse_headers,
            {cbase.Version.string: '1'},
            mock.ANY,
            mock.ANY)
        self.assertRaises(
            exc.HTTPNotAcceptable,
            cbase.Version.parse_headers,
            {cbase.Version.string: '1.2.3'},
            mock.ANY,
            mock.ANY)

    def test_parse_no_header(self):
        # this asserts that the minimum version string of "1.1" is applied
        version = cbase.Version.parse_headers({}, '1.1', '1.5')
        self.assertEqual((1, 1), version)

    def test_equals(self):
        ver_1 = cbase.Version(
            {cbase.Version.string: '123.456'}, mock.ANY, mock.ANY)
        ver_2 = cbase.Version(
            {cbase.Version.string: '123.456'}, mock.ANY, mock.ANY)
        ver_3 = cbase.Version(
            {cbase.Version.string: '654.321'}, mock.ANY, mock.ANY)
        self.assertTrue(hasattr(ver_1, '__eq__'))
        self.assertEqual(ver_1, ver_2)
        # Force __eq__ to be called and return False
        self.assertFalse(ver_1 == ver_3)  # noqa

    def test_not_equals(self):
        ver_1 = cbase.Version(
            {cbase.Version.string: '123.456'}, mock.ANY, mock.ANY)
        ver_2 = cbase.Version(
            {cbase.Version.string: '123.456'}, mock.ANY, mock.ANY)
        ver_3 = cbase.Version(
            {cbase.Version.string: '654.321'}, mock.ANY, mock.ANY)
        self.assertTrue(hasattr(ver_1, '__ne__'))
        self.assertNotEqual(ver_1, ver_3)
        # Force __ne__ to be called and return False
        self.assertFalse(ver_1 != ver_2)  # noqa

    def test_greaterthan(self):
        ver_1 = cbase.Version(
            {cbase.Version.string: '123.457'}, mock.ANY, mock.ANY)
        ver_2 = cbase.Version(
            {cbase.Version.string: '123.456'}, mock.ANY, mock.ANY)
        self.assertTrue(hasattr(ver_1, '__gt__'))
        self.assertGreater(ver_1, ver_2)
        # Force __gt__ to be called and return False
        self.assertFalse(ver_2 > ver_1)  # noqa

    def test_lessthan(self):
        # __lt__ is created by @functools.total_ordering, make sure it exists
        # and works
        ver_1 = cbase.Version(
            {cbase.Version.string: '123.456'}, mock.ANY, mock.ANY)
        ver_2 = cbase.Version(
            {cbase.Version.string: '123.457'}, mock.ANY, mock.ANY)
        self.assertTrue(hasattr(ver_1, '__lt__'))
        self.assertLess(ver_1, ver_2)
        # Force __lt__ to be called and return False
        self.assertFalse(ver_2 < ver_1)  # noqa



#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_log import log as logging

from ironic.common import metrics_utils
from ironic.conductor import periodics
from ironic import conf
from ironic.drivers.modules import graphical_console
from ironic.drivers.modules.redfish import utils as redfish_utils

LOG = logging.getLogger(__name__)
CONF = conf.CONF

METRICS = metrics_utils.get_metrics_logger(__name__)


class RedfishGraphicalConsole(graphical_console.GraphicalConsole):

    def get_app_name(self):
        return 'redfish-graphical'

    def get_properties(self):
        """Return the properties of the interface.

        :returns: dictionary of <property name>:<property description> entries.
        """
        return redfish_utils.COMMON_PROPERTIES.copy()

    def validate(self, task):
        """Validates the driver information needed by the redfish driver.

        :param task: a TaskManager instance containing the node to act on.
        :raises: InvalidParameterValue on malformed parameter(s)
        :raises: MissingParameterValue on missing parameter(s)
        """
        node = task.node
        redfish_utils.parse_driver_info(node)

    def get_app_info(self, task):
        """Information required by the app to connect to the console

        For redfish based consoles the app info will be the parsed driver
        info.

        :returns: dict containing parsed driver info
        """
        return redfish_utils.parse_driver_info(task.node)

    @METRICS.timer('RedfishGraphicalConsole._expire_console_sessions')
    @periodics.node_periodic(
        purpose='checking active console sessions',
        spacing=CONF.vnc.expire_console_session_interval,
        filters={'console_enabled': True},
        predicate_extra_fields=['console_interface', 'driver_internal_info'],
        predicate=lambda n: (
            'redfish-graphical' == n.console_interface
                and n.driver_internal_info.get('novnc_secret_token')
        ),
    )
    def _expire_redfish_console_sessions(self, task, manager, context):
        """Periodic task to close expired console sessions"""
        self._expire_console_sessions(task)



#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_config import cfg

from ironic.common.i18n import _
from ironic.common import keystone
from ironic.conf.api import Octal


CONF = cfg.CONF


opts = [
    cfg.StrOpt('auth_strategy',
               choices=[('noauth', _('no authentication')),
                        ('keystone', _('use the Identity service for '
                                       'authentication')),
                        ('http_basic', _('HTTP basic authentication'))],
               help=_('Authentication strategy used by JSON RPC. Defaults to '
                      'the global auth_strategy setting.')),
    cfg.StrOpt('http_basic_auth_user_file',
               default='/etc/ironic/htpasswd-json-rpc',
               help=_('Path to Apache format user authentication file used '
                      'when auth_strategy=http_basic')),
    cfg.HostAddressOpt('host_ip',
                       default='::',
                       help=_('The IP address or hostname on which JSON RPC '
                              'will listen.')),
    cfg.PortOpt('port',
                default=8089,
                help=_('The port to use for JSON RPC')),
    cfg.BoolOpt('use_ssl',
                default=False,
                help=_('Whether to use TLS for JSON RPC')),
    cfg.StrOpt('http_basic_username',
               deprecated_for_removal=True,
               deprecated_reason=_("Use username instead"),
               help=_("Name of the user to use for HTTP Basic authentication "
                      "client requests.")),
    cfg.StrOpt('http_basic_password',
               deprecated_for_removal=True,
               deprecated_reason=_("Use password instead"),
               secret=True,
               help=_("Password to use for HTTP Basic authentication "
                      "client requests.")),
    cfg.ListOpt('allowed_roles',
                default=['admin'],
                help=_("List of roles allowed to use JSON RPC")),
    cfg.StrOpt('unix_socket',
               help=_('Unix socket to listen on. Disables host_ip and port.')),
    cfg.Opt('unix_socket_mode', type=Octal(),
            help=_('File mode (an octal number) of the unix socket to '
                   'listen on. Ignored if unix_socket is not set.')),
]


def register_opts(conf):
    conf.register_opts(opts, group='json_rpc')
    keystone.register_auth_opts(conf, 'json_rpc')
    conf.set_default('timeout', 120, group='json_rpc')


def list_opts():
    return keystone.add_auth_opts(opts)


def auth_strategy():
    return CONF.json_rpc.auth_strategy or CONF.auth_strategy



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Create runbooks and runbook_steps tables

Revision ID: 66bd9c5604d5
Revises: 01f21d5e5195
Create Date: 2024-05-29 19:33:53.268794

"""

from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '66bd9c5604d5'
down_revision = '01f21d5e5195'


def upgrade():
    op.create_table(
        'runbooks',
        sa.Column('version', sa.String(length=15), nullable=True),
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.Column('id', sa.Integer(), nullable=False,
                  autoincrement=True),
        sa.Column('uuid', sa.String(length=36)),
        sa.Column('name', sa.String(length=255), nullable=False),
        sa.Column('disable_ramdisk', sa.Boolean, default=False),
        sa.Column('public', sa.Boolean, default=False),
        sa.Column('owner', sa.String(length=255), nullable=True),
        sa.Column('extra', sa.Text(), nullable=True),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint('uuid', name='uniq_runbooks0uuid'),
        sa.UniqueConstraint('name', name='uniq_runbooks0name'),
        mysql_engine='InnoDB',
        mysql_charset='UTF8MB3'
    )
    op.create_table(
        'runbook_steps',
        sa.Column('version', sa.String(length=15), nullable=True),
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.Column('id', sa.Integer(), nullable=False,
                  autoincrement=True),
        sa.Column('runbook_id', sa.Integer(), nullable=False,
                  autoincrement=False),
        sa.Column('interface', sa.String(length=255), nullable=False),
        sa.Column('step', sa.String(length=255), nullable=False),
        sa.Column('args', sa.Text, nullable=False),
        sa.Column('order', sa.Integer, nullable=False),
        sa.PrimaryKeyConstraint('id'),
        sa.ForeignKeyConstraint(['runbook_id'],
                                ['runbooks.id']),
        sa.Index('runbook_id', 'runbook_id'),
        sa.Index('runbook_steps_interface_idx', 'interface'),
        sa.Index('runbook_steps_step_idx', 'step'),
        mysql_engine='InnoDB',
        mysql_charset='UTF8MB3'
    )



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Tests for manipulating allocations via the DB API"""

from oslo_utils import uuidutils

from ironic.common import exception
from ironic.db import api as db_api
from ironic.tests.unit.db import base
from ironic.tests.unit.db import utils as db_utils


class AllocationsTestCase(base.DbTestCase):

    def setUp(self):
        super(AllocationsTestCase, self).setUp()
        self.node = db_utils.create_test_node()
        self.allocation = db_utils.create_test_allocation(name='host1')

    def test_create(self):
        dbapi = db_api.get_instance()
        allocation = dbapi.create_allocation({'resource_class': 'bm'})
        self.assertIsNotNone(allocation.uuid)
        self.assertEqual('allocating', allocation.state)

    def _create_test_allocation_range(self, count, start_idx=0, **kw):
        """Create the specified number of test allocation entries in DB

        It uses create_test_allocation method. And returns List of Allocation
        DB objects.

        :param count: Specifies the number of allocations to be created
        :returns: List of Allocation DB objects

        """
        return [db_utils.create_test_allocation(uuid=uuidutils.generate_uuid(),
                                                name='allocation' + str(i),
                                                **kw).uuid
                for i in range(start_idx, count + start_idx)]

    def test_get_allocation_by_id(self):
        res = self.dbapi.get_allocation_by_id(self.allocation.id)
        self.assertEqual(self.allocation.uuid, res.uuid)

    def test_get_allocation_by_id_that_does_not_exist(self):
        self.assertRaises(exception.AllocationNotFound,
                          self.dbapi.get_allocation_by_id, 99)

    def test_get_allocation_by_uuid(self):
        res = self.dbapi.get_allocation_by_uuid(self.allocation.uuid)
        self.assertEqual(self.allocation.id, res.id)

    def test_get_allocation_by_uuid_that_does_not_exist(self):
        self.assertRaises(exception.AllocationNotFound,
                          self.dbapi.get_allocation_by_uuid,
                          'EEEEEEEE-EEEE-EEEE-EEEE-EEEEEEEEEEEE')

    def test_get_allocation_by_name(self):
        res = self.dbapi.get_allocation_by_name(self.allocation.name)
        self.assertEqual(self.allocation.id, res.id)

    def test_get_allocation_by_name_that_does_not_exist(self):
        self.assertRaises(exception.AllocationNotFound,
                          self.dbapi.get_allocation_by_name, 'testfail')

    def test_get_allocation_list(self):
        uuids = self._create_test_allocation_range(6)
        # Also add the uuid for the allocation created in setUp()
        uuids.append(self.allocation.uuid)

        res = self.dbapi.get_allocation_list()
        self.assertEqual(set(uuids), {r.uuid for r in res})

    def test_get_allocation_list_sorted(self):
        uuids = self._create_test_allocation_range(6)
        # Also add the uuid for the allocation created in setUp()
        uuids.append(self.allocation.uuid)

        res = self.dbapi.get_allocation_list(sort_key='uuid')
        res_uuids = [r.uuid for r in res]
        self.assertEqual(sorted(uuids), res_uuids)

    def test_get_allocation_list_filter_by_state(self):
        self._create_test_allocation_range(6, state='error')

        res = self.dbapi.get_allocation_list(filters={'state': 'allocating'})
        self.assertEqual([self.allocation.uuid], [r.uuid for r in res])

        res = self.dbapi.get_allocation_list(filters={'state': 'error'})
        self.assertEqual(6, len(res))

    def test_get_allocation_list_filter_by_node(self):
        self._create_test_allocation_range(6)
        self.dbapi.update_allocation(self.allocation.id,
                                     {'node_id': self.node.id})

        res = self.dbapi.get_allocation_list(
            filters={'node_uuid': self.node.uuid})
        self.assertEqual([self.allocation.uuid], [r.uuid for r in res])

    def test_get_allocation_list_filter_by_rsc(self):
        self._create_test_allocation_range(6)
        self.dbapi.update_allocation(self.allocation.id,
                                     {'resource_class': 'very-large'})

        res = self.dbapi.get_allocation_list(
            filters={'resource_class': 'very-large'})
        self.assertEqual([self.allocation.uuid], [r.uuid for r in res])

    def test_get_allocation_list_filter_by_conductor_affinity(self):
        db_utils.create_test_conductor(id=1, hostname='host1')
        db_utils.create_test_conductor(id=2, hostname='host2')
        in_host1 = self._create_test_allocation_range(2, conductor_affinity=1)
        in_host2 = self._create_test_allocation_range(2, conductor_affinity=2,
                                                      start_idx=2)

        res = self.dbapi.get_allocation_list(
            filters={'conductor_affinity': 1})
        self.assertEqual(set(in_host1), {r.uuid for r in res})

        res = self.dbapi.get_allocation_list(
            filters={'conductor_affinity': 'host2'})
        self.assertEqual(set(in_host2), {r.uuid for r in res})

    def test_get_allocation_list_invalid_fields(self):
        self.assertRaises(exception.InvalidParameterValue,
                          self.dbapi.get_allocation_list, sort_key='foo')
        self.assertRaises(ValueError,
                          self.dbapi.get_allocation_list,
                          filters={'foo': 42})

    def test_destroy_allocation(self):
        self.dbapi.destroy_allocation(self.allocation.id)
        self.assertRaises(exception.AllocationNotFound,
                          self.dbapi.get_allocation_by_id, self.allocation.id)

    def test_destroy_allocation_with_node(self):
        self.dbapi.update_node(self.node.id,
                               {'allocation_id': self.allocation.id,
                                'instance_uuid': uuidutils.generate_uuid(),
                                'instance_info': {'traits': ['foo']}})
        self.dbapi.destroy_allocation(self.allocation.id)
        self.assertRaises(exception.AllocationNotFound,
                          self.dbapi.get_allocation_by_id, self.allocation.id)
        node = self.dbapi.get_node_by_id(self.node.id)
        self.assertIsNone(node.allocation_id)
        self.assertIsNone(node.instance_uuid)
        # NOTE(dtantsur): currently we do not clean up instance_info contents
        # on deallocation. It may be changed in the future.
        self.assertEqual(node.instance_info, {'traits': ['foo']})

    def test_destroy_allocation_that_does_not_exist(self):
        self.assertRaises(exception.AllocationNotFound,
                          self.dbapi.destroy_allocation, 99)

    def test_destroy_allocation_uuid(self):
        self.dbapi.destroy_allocation(self.allocation.uuid)

    def test_update_allocation(self):
        old_name = self.allocation.name
        new_name = 'newname'
        self.assertNotEqual(old_name, new_name)
        res = self.dbapi.update_allocation(self.allocation.id,
                                           {'name': new_name})
        self.assertEqual(new_name, res.name)

    def test_update_allocation_uuid(self):
        self.assertRaises(exception.InvalidParameterValue,
                          self.dbapi.update_allocation, self.allocation.id,
                          {'uuid': ''})

    def test_update_allocation_not_found(self):
        id_2 = 99
        self.assertNotEqual(self.allocation.id, id_2)
        self.assertRaises(exception.AllocationNotFound,
                          self.dbapi.update_allocation, id_2,
                          {'name': 'newname'})

    def test_update_allocation_duplicated_name(self):
        name1 = self.allocation.name
        allocation2 = db_utils.create_test_allocation(
            uuid=uuidutils.generate_uuid(), name='name2')
        self.assertRaises(exception.AllocationDuplicateName,
                          self.dbapi.update_allocation, allocation2.id,
                          {'name': name1})

    def test_update_allocation_with_node_id(self):
        res = self.dbapi.update_allocation(self.allocation.id,
                                           {'name': 'newname',
                                            'traits': ['foo'],
                                            'node_id': self.node.id})
        self.assertEqual('newname', res.name)
        self.assertEqual(['foo'], res.traits)
        self.assertEqual(self.node.id, res.node_id)

        node = self.dbapi.get_node_by_id(self.node.id)
        self.assertEqual(res.id, node.allocation_id)
        self.assertEqual(res.uuid, node.instance_uuid)
        self.assertEqual(['foo'], node.instance_info['traits'])

    def test_update_allocation_node_already_associated(self):
        existing_uuid = uuidutils.generate_uuid()
        self.dbapi.update_node(self.node.id, {'instance_uuid': existing_uuid})
        self.assertRaises(exception.NodeAssociated,
                          self.dbapi.update_allocation, self.allocation.id,
                          {'node_id': self.node.id, 'traits': ['foo']})

        # Make sure we do not see partial updates
        allocation = self.dbapi.get_allocation_by_id(self.allocation.id)
        self.assertEqual([], allocation.traits)
        self.assertIsNone(allocation.node_id)

        node = self.dbapi.get_node_by_id(self.node.id)
        self.assertIsNone(node.allocation_id)
        self.assertEqual(existing_uuid, node.instance_uuid)
        self.assertNotIn('traits', node.instance_info)

    def test_update_allocation_associated_with_another_node(self):
        db_utils.create_test_node(uuid=uuidutils.generate_uuid(),
                                  allocation_id=self.allocation.id,
                                  instance_uuid=self.allocation.uuid)

        self.assertRaises(exception.InstanceAssociated,
                          self.dbapi.update_allocation, self.allocation.id,
                          {'node_id': self.node.id, 'traits': ['foo']})

        # Make sure we do not see partial updates
        allocation = self.dbapi.get_allocation_by_id(self.allocation.id)
        self.assertEqual([], allocation.traits)
        self.assertIsNone(allocation.node_id)

        node = self.dbapi.get_node_by_id(self.node.id)
        self.assertIsNone(node.allocation_id)
        self.assertIsNone(node.instance_uuid)
        self.assertNotIn('traits', node.instance_info)

    def test_take_over_success(self):
        for i in range(2):
            db_utils.create_test_conductor(id=i, hostname='host-%d' % i)
        allocation = db_utils.create_test_allocation(conductor_affinity=0)

        self.assertTrue(self.dbapi.take_over_allocation(
            allocation.id, old_conductor_id=0, new_conductor_id=1))
        allocation = self.dbapi.get_allocation_by_id(allocation.id)
        self.assertEqual(1, allocation.conductor_affinity)

    def test_take_over_conflict(self):
        for i in range(3):
            db_utils.create_test_conductor(id=i, hostname='host-%d' % i)
        allocation = db_utils.create_test_allocation(conductor_affinity=2)

        self.assertFalse(self.dbapi.take_over_allocation(
            allocation.id, old_conductor_id=0, new_conductor_id=1))
        allocation = self.dbapi.get_allocation_by_id(allocation.id)
        # The affinity was not changed
        self.assertEqual(2, allocation.conductor_affinity)

    def test_take_over_allocation_not_found(self):
        self.assertRaises(exception.AllocationNotFound,
                          self.dbapi.take_over_allocation, 999, 0, 1)

    def test_create_allocation_duplicated_name(self):
        self.assertRaises(exception.AllocationDuplicateName,
                          db_utils.create_test_allocation,
                          uuid=uuidutils.generate_uuid(),
                          name=self.allocation.name)

    def test_create_allocation_duplicated_uuid(self):
        self.assertRaises(exception.AllocationAlreadyExists,
                          db_utils.create_test_allocation,
                          uuid=self.allocation.uuid)



# Copyright 2013 Hewlett-Packard Development Company, L.P.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import time
from unittest import mock

from oslo_config import cfg

from ironic.common import exception
from ironic.common import hash_ring
from ironic.common import utils
from ironic.tests.unit.db import base as db_base

CONF = cfg.CONF


class HashRingManagerTestCase(db_base.DbTestCase):

    use_groups = False

    def setUp(self):
        super(HashRingManagerTestCase, self).setUp()
        self.ring_manager = hash_ring.HashRingManager(
            use_groups=self.use_groups)

    def register_conductors(self):
        c1 = self.dbapi.register_conductor({
            'hostname': 'host1',
            'drivers': ['driver1', 'driver2'],
        })
        c2 = self.dbapi.register_conductor({
            'hostname': 'host2',
            'drivers': ['driver1'],
        })
        c3 = self.dbapi.register_conductor({
            'hostname': 'host3',
            'drivers': ['driver1, driver2'],
            'conductor_group': 'foogroup',
        })
        c4 = self.dbapi.register_conductor({
            'hostname': 'host4',
            'drivers': ['driver1'],
            'conductor_group': 'foogroup',
        })
        c5 = self.dbapi.register_conductor({
            'hostname': 'host5',
            'drivers': ['driver1'],
            'conductor_group': 'bargroup',
        })
        for c in (c1, c2, c3, c4, c5):
            self.dbapi.register_conductor_hardware_interfaces(
                c.id,
                [{'hardware_type': 'hardware-type', 'interface_type': 'deploy',
                  'interface_name': 'ansible', 'default': True},
                 {'hardware_type': 'hardware-type', 'interface_type': 'deploy',
                  'interface_name': 'direct', 'default': False}]
            )

    def test_hash_ring_manager_hardware_type_success(self):
        self.register_conductors()
        ring = self.ring_manager.get_ring('hardware-type', '')
        self.assertEqual(sorted(['host1', 'host2', 'host3', 'host4', 'host5']),
                         sorted(ring.nodes))

    def test_hash_ring_manager_hardware_type_success_groups(self):
        # groupings should be ignored here
        self.register_conductors()
        ring = self.ring_manager.get_ring('hardware-type', 'foogroup')
        self.assertEqual(sorted(['host1', 'host2', 'host3', 'host4', 'host5']),
                         sorted(ring.nodes))

    def test_hash_ring_manager_driver_not_found(self):
        self.register_conductors()
        self.assertRaises(exception.DriverNotFound,
                          self.ring_manager.get_ring,
                          'driver3', '')

    def test_hash_ring_manager_automatic_retry(self):
        self.assertRaises(exception.TemporaryFailure,
                          self.ring_manager.get_ring,
                          'hardware-type', '')
        self.register_conductors()
        self.ring_manager.get_ring('hardware-type', '')

    @mock.patch.object(utils, 'is_ironic_using_sqlite', autospec=True)
    def test_hash_ring_manager_reset_interval(self, is_sqlite_mock):
        is_sqlite_mock.return_value = False
        CONF.set_override('hash_ring_reset_interval', 30)
        # Just to simplify calculations
        CONF.set_override('hash_partition_exponent', 0)
        c1 = self.dbapi.register_conductor({
            'hostname': 'host1',
            'drivers': ['driver1', 'driver2'],
        })
        c2 = self.dbapi.register_conductor({
            'hostname': 'host2',
            'drivers': ['driver1'],
        })
        self.dbapi.register_conductor_hardware_interfaces(
            c1.id,
            [{'hardware_type': 'hardware-type', 'interface_type': 'deploy',
              'interface_name': 'ansible', 'default': True},
             {'hardware_type': 'hardware-type', 'interface_type': 'deploy',
              'interface_name': 'direct', 'default': False}]
        )

        ring = self.ring_manager.get_ring('hardware-type', '')
        self.assertEqual(1, len(ring))

        self.dbapi.register_conductor_hardware_interfaces(
            c2.id,
            [{'hardware_type': 'hardware-type', 'interface_type': 'deploy',
              'interface_name': 'ansible', 'default': True},
             {'hardware_type': 'hardware-type', 'interface_type': 'deploy',
              'interface_name': 'direct', 'default': False}]
        )
        ring = self.ring_manager.get_ring('hardware-type', '')
        # The new conductor is not known yet. Automatic retry does not kick in,
        # since there is an active conductor for the requested hardware type.
        self.assertEqual(1, len(ring))

        self.ring_manager.__class__._hash_rings = (
            self.ring_manager.__class__._hash_rings[0],
            time.monotonic() - 31
        )
        ring = self.ring_manager.get_ring('hardware-type', '')
        self.assertEqual(2, len(ring))
        self.assertEqual(3, is_sqlite_mock.call_count)

    @mock.patch.object(utils, 'is_ironic_using_sqlite', autospec=True)
    def test_hash_ring_manager_reset_interval_not_happen_sqlite(
            self, is_sqlite_mock):
        is_sqlite_mock.return_value = True
        CONF.set_override('hash_ring_reset_interval', 30)
        # Just to simplify calculations
        CONF.set_override('hash_partition_exponent', 0)
        c1 = self.dbapi.register_conductor({
            'hostname': 'host1',
            'drivers': ['driver1', 'driver2'],
        })
        c2 = self.dbapi.register_conductor({
            'hostname': 'host2',
            'drivers': ['driver1'],
        })
        self.dbapi.register_conductor_hardware_interfaces(
            c1.id,
            [{'hardware_type': 'hardware-type', 'interface_type': 'deploy',
              'interface_name': 'ansible', 'default': True},
             {'hardware_type': 'hardware-type', 'interface_type': 'deploy',
              'interface_name': 'direct', 'default': False}]
        )

        ring = self.ring_manager.get_ring('hardware-type', '')
        self.assertEqual(1, len(ring))

        self.dbapi.register_conductor_hardware_interfaces(
            c2.id,
            [{'hardware_type': 'hardware-type', 'interface_type': 'deploy',
              'interface_name': 'ansible', 'default': True},
             {'hardware_type': 'hardware-type', 'interface_type': 'deploy',
              'interface_name': 'direct', 'default': False}]
        )
        ring = self.ring_manager.get_ring('hardware-type', '')
        # The new conductor is not known yet. Automatic retry does not kick in,
        # since there is an active conductor for the requested hardware type.
        self.assertEqual(1, len(ring))

        self.ring_manager.__class__._hash_rings = (
            self.ring_manager.__class__._hash_rings[0],
            time.monotonic() - 31
        )
        ring = self.ring_manager.get_ring('hardware-type', '')
        self.assertEqual(1, len(ring))
        self.assertEqual(2, is_sqlite_mock.call_count)

    def test_hash_ring_manager_uncached(self):
        ring_mgr = hash_ring.HashRingManager(cache=False,
                                             use_groups=self.use_groups)
        ring = ring_mgr.ring
        self.assertIsNotNone(ring)
        self.assertEqual((None, 0), hash_ring.HashRingManager._hash_rings)


class HashRingManagerWithGroupsTestCase(HashRingManagerTestCase):

    use_groups = True

    def test_hash_ring_manager_hardware_type_success(self):
        self.register_conductors()
        ring = self.ring_manager.get_ring('hardware-type', '')
        self.assertEqual(sorted(['host1', 'host2']), sorted(ring.nodes))

    def test_hash_ring_manager_hardware_type_success_groups(self):
        self.register_conductors()
        ring = self.ring_manager.get_ring('hardware-type', 'foogroup')
        self.assertEqual(sorted(['host3', 'host4']), sorted(ring.nodes))
